"""Unit tests for the Beanstalk hosting service."""

from __future__ import unicode_literals

from reviewboard.hostingsvcs.testing import HostingServiceTestCase
from reviewboard.scmtools.crypto_utils import (decrypt_password,
                                               encrypt_password)


class BeanstalkTests(HostingServiceTestCase):
    """Unit tests for the Beanstalk hosting service."""

    service_name = 'beanstalk'
    fixtures = ['test_scmtools']

    default_account_data = {
        'password': encrypt_password(HostingServiceTestCase.default_password),
    }

    default_repository_extra_data = {
        'beanstalk_account_domain': 'mydomain',
        'beanstalk_repo_name': 'myrepo',
    }

    def test_service_support(self):
        """Testing Beanstalk service support capabilities"""
        self.assertFalse(self.service_class.supports_bug_trackers)
        self.assertTrue(self.service_class.supports_repositories)

    def test_get_repository_fields_for_git(self):
        """Testing Beanstalk.get_repository_fields for Git"""
        self.assertEqual(
            self.get_repository_fields(
                'Git',
                fields={
                    'beanstalk_account_domain': 'mydomain',
                    'beanstalk_repo_name': 'myrepo',
                }
            ),
            {
                'path': 'git@mydomain.beanstalkapp.com:/mydomain/myrepo.git',
                'mirror_path': ('https://mydomain.git.beanstalkapp.com/'
                                'myrepo.git'),
            })

    def test_get_repository_fields_for_subversion(self):
        """Testing Beanstalk.get_repository_fields for Subversion"""
        self.assertEqual(
            self.get_repository_fields(
                'Subversion',
                fields={
                    'beanstalk_account_domain': 'mydomain',
                    'beanstalk_repo_name': 'myrepo',
                }
            ),
            {
                'path': 'https://mydomain.svn.beanstalkapp.com/myrepo/',
            })

    def test_authorize(self):
        """Testing Beanstalk.authorize"""
        account = self.create_hosting_account(data={})
        service = account.service

        self.assertFalse(service.is_authorized())

        service.authorize('myuser', 'abc123', None)

        self.assertIn('password', account.data)
        self.assertNotEqual(account.data['password'], 'abc123')
        self.assertEqual(decrypt_password(account.data['password']), 'abc123')
        self.assertTrue(service.is_authorized())

    def test_check_repository(self):
        """Testing Beanstalk.check_repository"""
        with self.setup_http_test(payload=b'{}',
                                  expected_http_calls=1) as ctx:
            ctx.service.check_repository(beanstalk_account_domain='mydomain',
                                         beanstalk_repo_name='myrepo')

        ctx.assertHTTPCall(
            0,
            url=('https://mydomain.beanstalkapp.com/api/repositories/'
                 'myrepo.json'))

    def test_get_file_with_svn_and_base_commit_id(self):
        """Testing Beanstalk.get_file with Subversion and base commit ID"""
        self._test_get_file(
            tool_name='Subversion',
            revision='123',
            base_commit_id='456',
            expected_revision='123')

    def test_get_file_with_svn_and_revision(self):
        """Testing Beanstalk.get_file with Subversion and revision"""
        self._test_get_file(
            tool_name='Subversion',
            revision='123',
            base_commit_id=None,
            expected_revision='123')

    def test_get_file_with_git_and_base_commit_id(self):
        """Testing Beanstalk.get_file with Git and base commit ID"""
        self._test_get_file(
            tool_name='Git',
            revision='123',
            base_commit_id='456',
            expected_revision='123')

    def test_get_file_with_git_and_revision(self):
        """Testing Beanstalk.get_file with Git and revision"""
        self._test_get_file(
            tool_name='Git',
            revision='123',
            base_commit_id=None,
            expected_revision='123')

    def test_get_file_exists_with_svn_and_base_commit_id(self):
        """Testing Beanstalk.get_file_exists with Subversion and base commit ID
        """
        self._test_get_file_exists(
            tool_name='Subversion',
            revision='123',
            base_commit_id='456',
            expected_revision='123',
            expected_found=True)

    def test_get_file_exists_with_svn_and_revision(self):
        """Testing Beanstalk.get_file_exists with Subversion and revision"""
        self._test_get_file_exists(
            tool_name='Subversion',
            revision='123',
            base_commit_id=None,
            expected_revision='123',
            expected_found=True)

    def test_get_file_exists_with_git_and_base_commit_id(self):
        """Testing Beanstalk.get_file_exists with Git and base commit ID"""
        self._test_get_file_exists(
            tool_name='Git',
            revision='123',
            base_commit_id='456',
            expected_revision='456',
            expected_found=True)

    def test_get_file_exists_with_git_and_revision(self):
        """Testing Beanstalk.get_file_exists with Git and revision"""
        self._test_get_file_exists(
            tool_name='Git',
            revision='123',
            base_commit_id=None,
            expected_revision='123',
            expected_found=True)

    def _test_get_file(self, tool_name, revision, base_commit_id,
                       expected_revision):
        """Test file fetching.

        Args:
            tool_name (unicode):
                The name of the SCM Tool to test with.

            revision (unicode, optional):
                The revision to check.

            base_commit_id (unicode, optional):
                The base commit to fetch against.

            expected_revision (unicode, optional):
                The revision expected in the payload.
        """
        if tool_name == 'Git':
            payload = b'My data'
        else:
            payload = b'{"contents": "My data"}'

        with self.setup_http_test(payload=payload,
                                  expected_http_calls=1) as ctx:
            repository = ctx.create_repository(tool_name=tool_name)
            result = ctx.service.get_file(repository=repository,
                                          path='/path',
                                          revision=revision,
                                          base_commit_id=base_commit_id)

        self.assertIsInstance(result, bytes)
        self.assertEqual(result, b'My data')

        if tool_name == 'Git':
            expected_url = (
                'https://mydomain.beanstalkapp.com/api/repositories/'
                'myrepo/blob?id=%s&name=path'
                % expected_revision
            )
        else:
            expected_url = (
                'https://mydomain.beanstalkapp.com/api/repositories/'
                'myrepo/node.json?path=/path&revision=%s&contents=true'
                % expected_revision
            )

        ctx.assertHTTPCall(0, url=expected_url)

    def _test_get_file_exists(self, tool_name, revision, base_commit_id,
                              expected_revision, expected_found):
        """Test file existence checks.

        Args:
            tool_name (unicode):
                The name of the SCM Tool to test with.

            revision (unicode, optional):
                The revision to check.

            base_commit_id (unicode, optional):
                The base commit to fetch against.

            expected_revision (unicode, optional):
                The revision expected in the payload.

            expected_found (bool, optional):
                Whether a truthy response should be expected.
        """
        if expected_found:
            http_kwargs = {
                'payload': b'{}',
            }
        else:
            http_kwargs = {
                'status_code': 404,
            }

        with self.setup_http_test(expected_http_calls=1,
                                  **http_kwargs) as ctx:
            repository = ctx.create_repository(tool_name=tool_name)
            result = ctx.service.get_file_exists(repository=repository,
                                                 path='/path',
                                                 revision=revision,
                                                 base_commit_id=base_commit_id)

        self.assertEqual(result, expected_found)

        if not base_commit_id and tool_name == 'Git':
            expected_url = (
                'https://mydomain.beanstalkapp.com/api/repositories/myrepo/'
                'blob?id=%s&name=path'
                % expected_revision
            )
        else:
            expected_url = (
                'https://mydomain.beanstalkapp.com/api/repositories/myrepo/'
                'node.json?path=/path&revision=%s'
                % expected_revision
            )

        ctx.assertHTTPCall(0, url=expected_url)



from __future__ import unicode_literals

from django.core.urlresolvers import NoReverseMatch
from django.template.defaultfilters import date
from django.utils import six
from django.utils.html import (conditional_escape, escape, format_html,
                               format_html_join)
from django.utils.safestring import mark_safe
from django.utils.six.moves import reduce
from django.utils.translation import ugettext_lazy as _, ugettext
from djblets.datagrid.grids import CheckboxColumn, Column, DateTimeColumn
from djblets.siteconfig.models import SiteConfiguration

from reviewboard.accounts.models import Profile, ReviewRequestVisit
from reviewboard.avatars import avatar_services
from reviewboard.reviews.models import ReviewRequest
from reviewboard.reviews.templatetags.reviewtags import render_star
from reviewboard.site.urlresolvers import local_site_reverse


class BaseStarColumn(Column):
    """Indicates if an item is starred.

    This is the base class for all columns that deal with starring items.

    The star is interactive, allowing the user to star or unstar the item.
    """

    def __init__(self, *args, **kwargs):
        """Initialize the column."""
        super(BaseStarColumn, self).__init__(
            image_class='rb-icon rb-icon-star-on',
            image_alt=_('Starred'),
            detailed_label=_('Starred'),
            shrink=True,
            *args, **kwargs)

    def setup_state(self, state):
        """Set up the state for this column."""
        state.all_starred = set()

    def render_data(self, state, obj):
        """Return the rendered contents of the column."""
        obj.starred = obj.pk in state.all_starred
        return render_star(state.datagrid.request.user, obj)


class UsernameColumn(Column):
    """A column for showing a username and the user's avatar.

    The username and avatar will link to the user's profile page and will
    show basic profile information when hovering over the link.

    When constructing an instance of this column, the relation between the
    object being represented in the datagrid and the user can be specified
    as a tuple or list of field names forming a path to the user field.
    """

    AVATAR_SIZE = 24

    def __init__(self, label=_('Username'), user_relation=[], *args, **kwargs):
        """Initialize the column.

        Args:
            label (unicode, optional):
                The label for the column.

            user_relation (list of unicode, optional):
                A list of fields forming a relation path to the user. This can
                be left blank if representing the user.

            *args (tuple):
                Additional positional arguments to pass to the column.

            **kwargs (dict):
                Additional keyword arguments to pass to the column.
        """
        self._user_relation = user_relation

        super(UsernameColumn, self).__init__(
            label=label,
            db_field='__'.join(user_relation + ['username']),
            css_class='submitter-column',
            shrink=True,
            sortable=True,
            link=True,
            link_func=self._link_user,
            link_css_class='user',
            *args, **kwargs)

    def get_user(self, obj):
        """Return the user associated with this object.

        Args:
            obj (object):
                The object provided to the column.

        Returns:
            django.contrib.auth.models.User:
            The resulting user.
        """
        # Look up the user in the provided obj by traversing the relation.
        # If _user_relation is empty, then obj is the user.
        user = obj

        for field_name in self._user_relation:
            user = getattr(user, field_name)

        return user

    def render_data(self, state, obj):
        """Render the user's name and avatar as HTML.

        Args:
            state (djblets.datagrid.grids.StatefulColumn):
                The column state.

            obj (django.db.models.Model):
                The object being rendered in the datagrid.

        Returns:
            django.utils.safestring.SafeText:
            The HTML for the column.
        """
        user = self.get_user(obj)

        # If avatars are eanbled, we'll want to include that in the resulting
        # HTML.
        siteconfig = SiteConfiguration.objects.get_current()
        request = state.datagrid.request
        avatar_html = ''

        if siteconfig.get(avatar_services.AVATARS_ENABLED_KEY):
            avatar_service = avatar_services.for_user(user)

            if avatar_service:
                avatar_html = avatar_service.render(request=request,
                                                    user=user,
                                                    size=self.AVATAR_SIZE)

        # Render the link to the user page, using the avatar and username.
        username = user.username

        return format_html('{0}{1}', avatar_html, username)

    def augment_queryset(self, state, queryset):
        """Add additional queries to the queryset.

        This will select fields for the user and the user's profile, to
        help with query performance.

        Args:
            state (djblets.datagrid.grids.StatefulColumn):
                The column state.

            queryset (django.db.models.query.QuerySet):
                The queryset to augment.

        Returns:
            django.db.models.query.QuerySet:
            The resulting queryset.
        """
        user_field = '__'.join(self._user_relation)

        if user_field:
            fields = [user_field, '%s__profile' % user_field]
        else:
            fields = ['profile']

        return queryset.select_related(*fields)

    def _link_user(self, state, obj, *args):
        """Return the URL to link the user associated with this object.

        Args:
            state (djblets.datagrid.grids.StatefulColumn, unused):
                The column state.

            obj (object):
                The object provided to the column.

            *args (tuple):
                Additional keyword arguments provided to the method.

        Returns:
            unicode:
            The URL for the user.
        """
        return local_site_reverse(
            'user',
            request=state.datagrid.request,
            kwargs={
                'username': self.get_user(obj).username,
            })


class FullNameColumn(Column):
    """Shows the full name of the user when appropriate."""

    def augment_queryset(self, state, queryset):
        """Add additional queries to the queryset.

        This will select fields for the user and the user's profile, to
        help with query performance.

        Args:
            state (djblets.datagrid.grids.StatefulColumn):
                The column state.

            queryset (django.db.models.query.QuerySet):
                The queryset to augment.

        Returns:
            django.db.models.query.QuerySet:
            The resulting queryset.
        """
        return queryset.select_related('profile')

    def render_data(self, state, user):
        """Render the full name, or blank if not visible to the user.

        Args:
            state (djblets.datagrid.grids.StatefulColumn):
                The column state.

            user (django.contrib.auth.models.User):
                The user whose full name is to be rendered.

        Returns:
            unicode:
            Either the full name (if visible to the user) or an empty string.
        """
        profile = user.get_profile()

        if user.is_profile_visible(state.datagrid.request.user):
            display_name = \
                profile.get_display_name(state.datagrid.request.user)
        else:
            display_name = ''

        return escape(display_name)


class BugsColumn(Column):
    """Shows the list of bugs specified on a review request.

    The list of bugs will be linked to the bug tracker, if a bug tracker
    was configured for the repository the review request's change is on.
    """

    def __init__(self, *args, **kwargs):
        """Initialize the column."""
        # Note that we're enabling linking but overriding the link function
        # to return None. This is to disable the automatic linking to the
        # review request, so that the cell isn't generally clickable,
        # preventing visual and interaction issues with the bug links.
        super(BugsColumn, self).__init__(
            label=_('Bugs'),
            css_class='bugs',
            link=False,
            shrink=True,
            sortable=False,
            *args, **kwargs)

    def augment_queryset(self, state, queryset):
        """Add additional queries to the queryset."""
        return queryset.select_related('repository')

    def render_data(self, state, review_request):
        """Return the rendered contents of the column."""
        bugs = review_request.get_bug_list()
        repository = review_request.repository
        local_site_name = None

        if review_request.local_site:
            local_site_name = review_request.local_site.name

        if repository and repository.bug_tracker:
            links = []

            for bug in bugs:
                try:
                    url = local_site_reverse(
                        'bug_url',
                        local_site_name=local_site_name,
                        args=[review_request.display_id, bug])
                    links.append(
                        format_html('<a class="bug" href="{0}">{1}</a>',
                                    url, bug))
                except NoReverseMatch:
                    links.append(escape(bug))

            return ', '.join(links)

        return format_html_join(
            ', ',
            '<span class="bug">{0}</span>',
            ((bug,) for bug in bugs)
        )


class ReviewRequestCheckboxColumn(CheckboxColumn):
    """A column containing a check-box."""

    def render_data(self, state, obj):
        """Return the rendered contents of the column."""
        if self.is_selectable(state, obj):
            checked = ''

            if self.is_selected(state, obj):
                checked = 'checked="true"'

            return ('<input type="checkbox" data-object-id="%s" '
                    'data-checkbox-name="%s" %s />'
                    % (obj.display_id, escape(self.checkbox_name), checked))
        else:
            return ''


class DateTimeSinceColumn(DateTimeColumn):
    """Displays how long it has been since a given date/time.

    These columns will dynamically update as the page is shown, so that the
    number of minutes, hours, days, etc. ago is correct.
    """

    def render_data(self, state, obj):
        """Return the rendered contents of the column."""
        return '<time class="timesince" datetime="%s">%s</time>' % (
            date(getattr(obj, self.field_name), 'c'),
            super(DateTimeSinceColumn, self).render_data(state, obj))


class DiffUpdatedColumn(DateTimeColumn):
    """Shows the date/time that the diff was last updated."""

    def __init__(self, *args, **kwargs):
        """Initialize the column."""
        super(DiffUpdatedColumn, self).__init__(
            label=_('Diff Updated'),
            db_field='diffset_history__last_diff_updated',
            field_name='last_diff_updated',
            sortable=True,
            link=False,
            *args, **kwargs)

    def augment_queryset(self, state, queryset):
        """Add additional queries to the queryset."""
        return queryset.select_related('diffset_history')

    def render_data(self, state, obj):
        """Return the rendered contents of the column."""
        if obj.diffset_history.last_diff_updated:
            return super(DiffUpdatedColumn, self).render_data(
                state, obj.diffset_history)
        else:
            return ''


class DiffUpdatedSinceColumn(DateTimeSinceColumn):
    """Shows the elapsed time since the diff was last updated."""

    def __init__(self, *args, **kwargs):
        """Initialize the column."""
        super(DiffUpdatedSinceColumn, self).__init__(
            label=_('Diff Updated'),
            db_field='diffset_history__last_diff_updated',
            field_name='last_diff_updated',
            sortable=True,
            link=False,
            *args, **kwargs)

    def augment_queryset(self, state, queryset):
        """Add additional queries to the queryset."""
        return queryset.select_related('diffset_history')

    def render_data(self, state, obj):
        """Return the rendered contents of the column."""
        if obj.diffset_history.last_diff_updated:
            return super(DiffUpdatedSinceColumn, self).render_data(
                state, obj.diffset_history)
        else:
            return ''


class GroupMemberCountColumn(Column):
    """Shows the number of users that are part of a review group."""

    def __init__(self, *args, **kwargs):
        """Initialize the column."""
        super(GroupMemberCountColumn, self).__init__(
            link=True,
            link_func=self.link_to_object,
            *args, **kwargs)

    def render_data(self, state, group):
        """Return the rendered contents of the column."""
        return six.text_type(group.users.count())

    def link_to_object(self, state, group, value):
        """Return the link to the object in the column."""
        return local_site_reverse('group-members',
                                  request=state.datagrid.request,
                                  args=[group.name])


class GroupsColumn(Column):
    """Shows the list of groups requested to review the review request."""

    def __init__(self, *args, **kwargs):
        """Initialize the column."""
        super(GroupsColumn, self).__init__(
            label=_('Groups'),
            detailed_label=_('Target Groups'),
            sortable=False,
            shrink=False,
            *args, **kwargs)

    def augment_queryset(self, state, queryset):
        """Add additional queries to the queryset."""
        return queryset.prefetch_related('target_groups')

    def render_data(self, state, review_request):
        """Return the rendered contents of the column."""
        groups = review_request.target_groups.all()
        return reduce(lambda a, d: a + d.name + ' ', groups, '')


class MyCommentsColumn(Column):
    """Shows if the current user has reviewed the review request."""

    def __init__(self, *args, **kwargs):
        """Initialize the column."""
        super(MyCommentsColumn, self).__init__(
            image_class='rb-icon rb-icon-datagrid-comment-draft',
            image_alt=_('My Comments'),
            detailed_label=_('My Comments'),
            shrink=True,
            *args, **kwargs)

        # XXX It'd be nice to be able to sort on this, but datagrids currently
        # can only sort based on stored (in the DB) values, not computed
        # values.

    def augment_queryset(self, state, queryset):
        """Add additional queries to the queryset."""
        user = state.datagrid.request.user

        if user.is_anonymous():
            return queryset

        query_dict = {
            'user_id': six.text_type(user.id),
        }

        return queryset.extra(select={
            'mycomments_my_reviews': """
                SELECT COUNT(1)
                  FROM reviews_review
                  WHERE reviews_review.user_id = %(user_id)s
                    AND reviews_review.review_request_id =
                        reviews_reviewrequest.id
            """ % query_dict,
            'mycomments_private_reviews': """
                SELECT COUNT(1)
                  FROM reviews_review
                  WHERE reviews_review.user_id = %(user_id)s
                    AND reviews_review.review_request_id =
                        reviews_reviewrequest.id
                    AND NOT reviews_review.public
            """ % query_dict,
            'mycomments_shipit_reviews': """
                SELECT COUNT(1)
                  FROM reviews_review
                  WHERE reviews_review.user_id = %(user_id)s
                    AND reviews_review.review_request_id =
                        reviews_reviewrequest.id
                    AND reviews_review.ship_it
            """ % query_dict,
        })

    def render_data(self, state, review_request):
        """Return the rendered contents of the column."""
        user = state.datagrid.request.user

        if user.is_anonymous() or review_request.mycomments_my_reviews == 0:
            return ''

        # Priority is ranked in the following order:
        #
        # 1) Non-public (draft) reviews
        # 2) Public reviews marked "Ship It"
        # 3) Public reviews not marked "Ship It"
        if review_request.mycomments_private_reviews > 0:
            icon_class = 'rb-icon-datagrid-comment-draft'
            image_alt = _('Comments drafted')
        else:
            if review_request.mycomments_shipit_reviews > 0:
                icon_class = 'rb-icon-datagrid-comment-shipit'
                image_alt = _('Comments published. Ship it!')
            else:
                icon_class = 'rb-icon-datagrid-comment'
                image_alt = _('Comments published')

        return '<div class="rb-icon %s" title="%s"></div>' % \
               (icon_class, image_alt)


class NewUpdatesColumn(Column):
    """Indicates if there are new updates on a review request.

    This will show an icon if the review request has had any new updates
    or reviews since the user last saw it.
    """

    def __init__(self, *args, **kwargs):
        """Initialize the column."""
        super(NewUpdatesColumn, self).__init__(
            image_class='rb-icon rb-icon-new-updates',
            image_alt=_('New Updates'),
            detailed_label=_('New Updates'),
            shrink=True,
            *args, **kwargs)

    def render_data(self, state, review_request):
        """Return the rendered contents of the column."""

        # Review requests for un-authenticated users will not contain the
        # new_review_count attribute, so confirm its existence before
        # attempting to access.
        if (hasattr(review_request, 'new_review_count') and
            review_request.new_review_count > 0):
            return '<div class="%s" title="%s" />' % \
                   (self.image_class, self.image_alt)

        return ''


class PendingCountColumn(Column):
    """Shows the pending number of review requests for a user or group.

    This will show the pending number of review requests for the given
    review group or user. It only applies to group or user lists.
    """

    def render_data(self, state, obj):
        """Return the rendered contents of the column."""
        return six.text_type(
            getattr(obj, self.field_name).filter(
                public=True, status='P').count())


class PeopleColumn(Column):
    """Shows the list of people requested to review the review request."""

    def __init__(self, *args, **kwargs):
        """Initialize the column."""
        super(PeopleColumn, self).__init__(
            label=_('People'),
            detailed_label=_('Target People'),
            sortable=False,
            shrink=False,
            *args, **kwargs)

    def augment_queryset(self, state, queryset):
        """Add additional queries to the queryset."""
        return queryset.prefetch_related('target_people')

    def render_data(self, state, review_request):
        """Return the rendered contents of the column."""
        people = review_request.target_people.all()
        return reduce(lambda a, d: a + d.username + ' ', people, '')


class RepositoryColumn(Column):
    """Shows the name of the repository the review request's change is on."""

    def __init__(self, *args, **kwargs):
        """Initialize the column."""
        super(RepositoryColumn, self).__init__(
            label=_('Repository'),
            db_field='repository__name',
            shrink=True,
            sortable=True,
            link=False,
            css_class='repository-column',
            *args, **kwargs)

    def augment_queryset(self, state, queryset):
        """Add additional queries to the queryset."""
        return queryset.select_related('repository')

    def render_data(self, state, obj):
        """Return the rendered contents of the column."""
        return super(RepositoryColumn, self).render_data(state, obj) or ''


class ReviewCountColumn(Column):
    """Shows the number of published reviews for a review request."""

    def __init__(self, *args, **kwargs):
        """Initialize the column."""
        super(ReviewCountColumn, self).__init__(
            label=_('Reviews'),
            detailed_label=_('Number of Reviews'),
            shrink=True,
            link=True,
            link_func=self.link_to_object,
            *kwargs, **kwargs)

    def render_data(self, state, review_request):
        """Return the rendered contents of the column."""
        return six.text_type(review_request.publicreviewcount_count)

    def augment_queryset(self, state, queryset):
        """Add additional queries to the queryset."""
        return queryset.extra(select={
            'publicreviewcount_count': """
                SELECT COUNT(*)
                  FROM reviews_review
                  WHERE reviews_review.public
                    AND reviews_review.base_reply_to_id is NULL
                    AND reviews_review.review_request_id =
                        reviews_reviewrequest.id
            """
        })

    def link_to_object(self, state, review_request, value):
        """Return the link to the object in the column."""
        return '%s#last-review' % review_request.get_absolute_url()


class ReviewGroupStarColumn(BaseStarColumn):
    """Indicates if a review group is starred.

    The star is interactive, allowing the user to star or unstar the group.
    """

    def augment_queryset(self, state, queryset):
        """Add additional queries to the queryset."""
        user = state.datagrid.request.user

        if user.is_anonymous():
            return queryset

        try:
            profile = user.get_profile()
        except Profile.DoesNotExist:
            return queryset

        state.all_starred = set(
            profile.starred_groups.filter(
                pk__in=state.datagrid.id_list).values_list('pk', flat=True))

        return queryset


class ReviewRequestIDColumn(Column):
    """Displays the ID of the review request."""

    def __init__(self, *args, **kwargs):
        """Initialize the column."""
        super(ReviewRequestIDColumn, self).__init__(
            label=_('ID'),
            detailed_label=_('Review Request ID'),
            shrink=True,
            link=True,
            sortable=True,
            *args, **kwargs)

    def get_sort_field(self, state):
        """Return the model field for sorting this column."""
        if state.datagrid.local_site:
            return 'local_id'
        else:
            return 'id'

    def render_data(self, state, review_request):
        """Return the rendered contents of the column."""
        return review_request.display_id


class ReviewRequestStarColumn(BaseStarColumn):
    """Indicates if a review request is starred.

    The star is interactive, allowing the user to star or unstar the
    review request.
    """

    def augment_queryset(self, state, queryset):
        """Add additional queries to the queryset."""
        user = state.datagrid.request.user

        if user.is_anonymous():
            return queryset

        try:
            profile = user.get_profile()
        except Profile.DoesNotExist:
            return queryset

        state.all_starred = set(
            profile.starred_review_requests.filter(
                pk__in=state.datagrid.id_list).values_list('pk', flat=True))

        return queryset


class ShipItColumn(Column):
    """Shows the "Ship It" count for a review request."""

    def __init__(self, *args, **kwargs):
        """Initialize the column."""
        super(ShipItColumn, self).__init__(
            image_class='rb-icon rb-icon-datagrid-column-shipits-issues',
            image_alt=_('Ship It!/Issue Counts'),
            detailed_label=_('Ship It!/Issue Counts'),
            db_field='shipit_count',
            sortable=True,
            shrink=True,
            *args, **kwargs)

    def render_data(self, state, review_request):
        """Return the rendered contents of the column.

        Args:
            state (djblets.datagrid.grids.StatefulColumn):
                The state for the datagrid.

            review_request (reviewboard.reviews.models.review_request.
                            ReviewRequest):
                The review request.

        Returns:
            django.utils.safestring.SafeText:
            The rendered HTML for the column.
        """
        open_issues = review_request.issue_open_count
        verifying_issues = review_request.issue_verifying_count

        if open_issues > 0 and verifying_issues > 0:
            return self._render_counts([
                {
                    'count': open_issues,
                    'title': _('Open issue count'),
                },
                {
                    'count': verifying_issues,
                    'css_class': 'issue-verifying-count',
                    'icon_name': 'issue-verifying',
                    'title': _('Verifying issue count'),
                },
            ])
        elif open_issues > 0:
            return self._render_counts([{
                'count': open_issues,
                'title': _('Open issue count'),
            }])
        elif verifying_issues > 0:
            return self._render_counts([{
                'count': verifying_issues,
                'icon_name': 'issue-verifying',
                'title': _('Verifying issue count'),
            }])
        elif review_request.shipit_count:
            return self._render_counts(
                [{
                    'count': review_request.shipit_count,
                    'css_class': 'shipit-count',
                    'icon_name': 'shipit',
                    'title': _('Ship It! count'),
                }],
                container_css_class='shipit-count-container')
        else:
            return ''

    def _render_counts(self, count_details,
                       container_css_class='issue-count-container'):
        """Render the counts for the column.

        This will render a container bubble in the column and render each
        provided count and icon in the bubble. This can be used for issues,
        Ship Its, or anything else we need down the road.

        Args:
            count_details (list of dict):
                The list of details for the count. This must have ``count``
                and ``title`` keys, and may optionally have ``css_class`` and
                ``icon_name`` keys.

            container_css_class (unicode, optional):
                The optional CSS class name for the outer container.

        Returns:
            django.utils.safestring.SafeText:
            The resulting HTML for the counts bubble.
        """
        # Note that the HTML is very whitespace-sensitive, so don't try to
        # change the templates to be nicely indented. The spacing is this way
        # for a reason.
        #
        # We also can't use format_html_join, unfortunately, as that doesn't
        # support keyword arguments.
        return format_html(
            '<div class="{container_css_class}">{count_html}</div>',
            container_css_class=container_css_class,
            count_html=mark_safe(''.join(
                format_html(
                    '<span class="{css_class}">'
                    '<span class="rb-icon rb-icon-datagrid-{icon_name}"'
                    '      title="{title}"></span>'
                    '{count}'
                    '</span>',
                    **dict({
                        'css_class': 'issue-count',
                        'icon_name': 'open-issues',
                    }, **count_detail))
                for count_detail in count_details
            )))


class SummaryColumn(Column):
    """Shows the summary of a review request.

    This will also prepend the draft/submitted/discarded state, if any,
    to the summary.
    """

    def __init__(self, *args, **kwargs):
        """Initialize the column."""
        super(SummaryColumn, self).__init__(
            label=_('Summary'),
            expand=True,
            link=True,
            link_css_class='review-request-link',
            css_class='summary',
            sortable=True,
            *args, **kwargs)

    def augment_queryset(self, state, queryset):
        """Add additional queries to the queryset."""
        user = state.datagrid.request.user

        if user.is_anonymous():
            return queryset

        return queryset.extra(select={
            'draft_summary': """
                SELECT reviews_reviewrequestdraft.summary
                  FROM reviews_reviewrequestdraft
                  WHERE reviews_reviewrequestdraft.review_request_id =
                        reviews_reviewrequest.id
            """,
            'visibility': """
                SELECT accounts_reviewrequestvisit.visibility
                  FROM accounts_reviewrequestvisit
                 WHERE accounts_reviewrequestvisit.review_request_id =
                       reviews_reviewrequest.id
                   AND accounts_reviewrequestvisit.user_id = %(user_id)s
            """ % {
                'user_id': six.text_type(user.id)
            }
        })

    def render_data(self, state, review_request):
        """Return the rendered contents of the column.

        Args:
            state (djblets.datagrids.grids.StatefulColumn):
                The state for the datagrid.

            review_request (reviewboard.reviews.models.review_request.ReviewRequest):
                The review request.

        Returns:
            django.utils.safestring.SafeText:
            The rendered column.
        """
        summary = review_request.summary
        labels = []

        if review_request.submitter_id == state.datagrid.request.user.id:
            if review_request.draft_summary is not None:
                summary = review_request.draft_summary
                labels.append(('label-draft', _('Draft')))
            elif (not review_request.public and
                  review_request.status == ReviewRequest.PENDING_REVIEW):
                labels.append(('label-draft', _('Draft')))

        # review_request.visibility is not defined when the user is not
        # logged in.
        if state.datagrid.request.user.is_authenticated():
            if review_request.visibility == ReviewRequestVisit.ARCHIVED:
                labels.append(('label-archived', _('Archived')))
            elif review_request.visibility == ReviewRequestVisit.MUTED:
                labels.append(('label-muted', _('Muted')))

        if review_request.status == ReviewRequest.SUBMITTED:
            labels.append(('label-submitted', _('Submitted')))
        elif review_request.status == ReviewRequest.DISCARDED:
            labels.append(('label-discarded', _('Discarded')))

        result = [
            format_html_join('', '<label class="{0}">{1}</label>', labels)
        ]

        if summary:
            result.append(format_html('<span>{0}</span>', summary))
        else:
            result.append(format_html('<span class="no-summary">{0}</span>',
                                      _('No Summary')))

        return mark_safe(''.join(result))


class ReviewSummaryColumn(SummaryColumn):
    """Shows the summary of the review request of a review.

    This does not (yet) prepend the draft/submitted/discarded state, if any,
    to the summary.
    """

    def __init__(self, *args, **kwargs):
        """Initialize the column."""
        super(SummaryColumn, self).__init__(
            label=_('Review Request Summary'),
            expand=True,
            link=True,
            css_class='summary',
            *args, **kwargs)

    def render_data(self, state, review):
        """Return the rendered contents of the column."""
        return conditional_escape(review.review_request.summary)

    def augment_queryset(self, state, queryset):
        """Add additional queries to the queryset."""
        return queryset.select_related('review_request')


class ToMeColumn(Column):
    """Indicates if the user is requested to review the change.

    This will show an indicator if the user is on the Target People reviewers
    list.
    """

    def __init__(self, *args, **kwargs):
        """Initialize the column."""
        raquo = '\u00BB'

        super(ToMeColumn, self).__init__(
            label=raquo,
            detailed_label=_('To Me'),
            detailed_label_html=(ugettext('%s To Me') % raquo),
            shrink=True,
            *args, **kwargs)

    def augment_queryset(self, state, queryset):
        """Add additional queries to the queryset."""
        user = state.datagrid.request.user

        if user.is_authenticated():
            state.all_to_me = set(
                user.directed_review_requests.filter(
                    pk__in=state.datagrid.id_list).values_list('pk',
                                                               flat=True))
        else:
            state.all_to_me = set()

        return queryset

    def render_data(self, state, review_request):
        """Return the rendered contents of the column."""
        if review_request.pk in state.all_to_me:
            return ('<div title="%s"><b>&raquo;</b></div>'
                    % (self.detailed_label))

        return ''


class DiffSizeColumn(Column):
    """Indicates line add/delete counts for the latest diffset."""

    def __init__(self, *args, **kwargs):
        """Initialize the column."""
        super(DiffSizeColumn, self).__init__(
            label=_('Diff Size'),
            sortable=False,
            shrink=True,
            *args, **kwargs)

    def render_data(self, state, review_request):
        """Return the rendered contents of the column."""
        if review_request.repository_id is None:
            return ''

        diffsets = list(review_request.diffset_history.diffsets.all())

        if not diffsets:
            return ''

        diffset = diffsets[-1]

        counts = diffset.get_total_line_counts()
        insert_count = counts.get('raw_insert_count')
        delete_count = counts.get('raw_delete_count')
        result = []

        if insert_count:
            result.append('<span class="diff-size-column insert">+%d</span>' %
                          insert_count)

        if delete_count:
            result.append('<span class="diff-size-column delete">-%d</span>' %
                          delete_count)

        if result:
            return '&nbsp;'.join(result)

        return ''

    def augment_queryset(self, state, queryset):
        """Add additional queries to the queryset.

        This will prefetch the diffsets and filediffs needed to perform the
        line calculations.

        Args:
            state (djblets.datagrid.grids.StatefulColumn):
                The column state.

            queryset (django.db.models.query.QuerySet):
                The queryset to augment.

        Returns:
            django.db.models.query.QuerySet:
            The resulting queryset.
        """
        # TODO: Update this to fetch only the specific fields when we move
        #       to a newer version of Django.
        return queryset.prefetch_related('diffset_history__diffsets',
                                         'diffset_history__diffsets__files')



from __future__ import unicode_literals

import markdown
from django import template
from django.utils.safestring import mark_safe
from djblets.markdown import markdown_unescape


register = template.Library()


@register.filter
def markdown_email_html(text, is_rich_text):
    if not is_rich_text:
        return text

    # We use XHTML1 instead of HTML5 to ensure the results can be parsed by
    # an XML parser. This is actually needed for the main Markdown renderer
    # for the web UI, but consistency is good here.
    return mark_safe(markdown.markdown(
        text,
        output_format='xhtml1',
        extensions=[
            'markdown.extensions.fenced_code',
            'markdown.extensions.codehilite',
            'markdown.extensions.tables',
            'markdown.extensions.sane_lists',
            'markdown.extensions.smart_strong',
            'pymdownx.tilde',
            'djblets.markdown.extensions.escape_html',
            'djblets.markdown.extensions.wysiwyg_email',
        ],
        extension_configs={
            'codehilite': {
                'noclasses': True,
            },
        }))


@register.filter
def markdown_email_text(text, is_rich_text):
    if not is_rich_text:
        return text

    return markdown_unescape(text)



"""A module exposing the search backends registry."""

from __future__ import unicode_literals

from reviewboard.search.search_backends.registry import SearchBackendRegistry


search_backend_registry = SearchBackendRegistry()



"""Models for OAuth2 applications."""

from __future__ import unicode_literals

from django.contrib.auth.models import User
from django.db import models
from django.utils.translation import ugettext_lazy as _
from djblets.db.fields import JSONField
from oauth2_provider.models import AbstractApplication

from reviewboard.site.models import LocalSite


class Application(AbstractApplication):
    """An OAuth2 application.

    This model is specialized so that it can be limited to a
    :py:class:`~reviewboard.site.models.LocalSite`.
    """

    enabled = models.BooleanField(
        verbose_name=_('Enabled'),
        help_text=_('Whether or not this application can be used to '
                    'authenticate with Review Board.'),
        default=True,
    )

    original_user = models.ForeignKey(
        verbose_name=_('Original User'),
        to=User,
        blank=True,
        null=True,
        help_text=_('The original owner of this application.')
    )

    local_site = models.ForeignKey(
        verbose_name=_('Local Site'),
        to=LocalSite,
        related_name='oauth_applications',
        blank=True,
        null=True,
        help_text=_('An optional Local Site to limit this application to.<br>'
                    'If specified, only users with access to the Local Site '
                    'will be able to use the application.'),
    )

    extra_data = JSONField(
        _('Extra Data'),
        null=True,
        default=dict,
    )

    @property
    def is_disabled_for_security(self):
        """Whether or not this application is disabled for security reasons.

        This will be ``True`` when the :py:attr:`original_owner` no longer
        has access to the :py:attr:`local_site` this application is associated
        with.
        """
        return not self.enabled and self.original_user_id is not None

    def clean(self):
        """Validate the application.

        We do the validation for this in :py:meth:`ApplicationForm.clean()
        <reviewboard.oauth.forms.ApplicationForm.clean` so that we can have
        errors for ``authorization_grant_type`` and ``redirect_uris`` conflicts
        show up on the appropriate field. The parent class does the same
        validation, but as a result it will have form-wide errors instead of
        per-field errors for the above two fields when they are in conflict.
        Therefore we avoid that validation by making this a no-op.
        """
        pass

    def is_accessible_by(self, user, local_site=None):
        """Return whether or not the user has access to this Application.

        A user has access if one of the following conditions is met:

        * The user owns the Application.
        * The user is an administrator.
        * The user is a Local Site administrator on the Local Site the
          Application is assigned to.

        Args:
            user (django.contrib.auth.models.User):
                The user in question.

            local_site (reviewboard.site.models.LocalSite):
                The Local Site the user would access this Application under.

        Returns:
            bool:
            Whether or not the given user has access to information about
            this Application.
        """
        return (user.is_authenticated() and
                (self.user == user or
                 user.is_superuser or
                 (self.local_site_id is not None and
                  local_site is not None and
                  self.local_site_id == local_site.pk and
                  local_site.is_mutable_by(user))))

    def is_mutable_by(self, user, local_site=None):
        """Return whether or not the user can modify this Application.

        A user has access if one of the following conditions is met:

        * The user owns the Application.
        * The user is an administrator.
        * The user is a Local Site administrator on the Local Site the
          Application is assigned to.

        Args:
            user (django.contrib.auth.models.User):
                The user in question.

            local_site (reviewboard.site.models.LocalSite):
                The Local Site the user would modify this Application under.

        Returns:
            bool:
            Whether or not the given user can modify this Application.
        """
        return self.is_accessible_by(user, local_site=local_site)

    class Meta:
        db_table = 'reviewboard_oauth_application'
        verbose_name = _('OAuth Application')
        verbose_name_plural = _('OAuth Applications')



from __future__ import unicode_literals

from collections import deque

from django.template.loader import render_to_string

from reviewboard.reviews.errors import DepthLimitExceededError


#: The maximum depth limit of any action instance.
MAX_DEPTH_LIMIT = 2

#: The mapping of all action IDs to their corresponding action instances.
_all_actions = {}

#: All top-level action IDs (in their left-to-right order of appearance).
_top_level_ids = deque()

#: Determines if the default action instances have been populated yet.
_populated = False


class BaseReviewRequestAction(object):
    """A base class for an action that can be applied to a review request.

    Creating an action requires subclassing :py:class:`BaseReviewRequestAction`
    and overriding any fields/methods as desired. Different instances of the
    same subclass can also override the class fields with their own instance
    fields.

    Example:
        .. code-block:: python

           class UsedOnceAction(BaseReviewRequestAction):
               action_id = 'once'
               label = 'This is used once.'

           class UsedMultipleAction(BaseReviewRequestAction):
               def __init__(self, action_id, label):
                   super(UsedMultipleAction, self).__init__()

                   self.action_id = 'repeat-' + action_id
                   self.label = 'This is used multiple times,'

    Note:
        Since the same action will be rendered for multiple different users in
        a multithreaded environment, the action state should not be modified
        after initialization. If we want different action attributes at
        runtime, then we can override one of the getter methods (such as
        :py:meth:`get_label`), which by default will simply return the original
        attribute from initialization.
    """

    #: The ID of this action. Must be unique across all types of actions and
    #: menu actions, at any depth.
    action_id = None

    #: The label that displays this action to the user.
    label = None

    #: The URL to invoke if this action is clicked.
    url = '#'

    #: Determines if this action should be initially hidden to the user.
    hidden = False

    def __init__(self):
        """Initialize this action.

        By default, actions are top-level and have no children.
        """
        self._parent = None
        self._max_depth = 0

    def copy_to_dict(self, context):
        """Copy this action instance to a dictionary.

        Args:
            context (django.template.Context):
                The collection of key-value pairs from the template.

        Returns:
            dict: The corresponding dictionary.
        """
        return {
            'action_id': self.action_id,
            'label': self.get_label(context),
            'url': self.get_url(context),
            'hidden': self.get_hidden(context),
        }

    def get_label(self, context):
        """Return this action's label.

        Args:
            context (django.template.Context):
                The collection of key-value pairs from the template.

        Returns:
            unicode: The label that displays this action to the user.
        """
        return self.label

    def get_url(self, context):
        """Return this action's URL.

        Args:
            context (django.template.Context):
                The collection of key-value pairs from the template.

        Returns:
            unicode: The URL to invoke if this action is clicked.
        """
        return self.url

    def get_hidden(self, context):
        """Return whether this action should be initially hidden to the user.

        Args:
            context (django.template.Context):
                The collection of key-value pairs from the template.

        Returns:
            bool: Whether this action should be initially hidden to the user.
        """
        return self.hidden

    def should_render(self, context):
        """Return whether or not this action should render.

        The default implementation is to always render the action everywhere.

        Args:
            context (django.template.Context):
                The collection of key-value pairs available in the template
                just before this action is to be rendered.

        Returns:
            bool: Determines if this action should render.
        """
        return True

    @property
    def max_depth(self):
        """Lazily compute the max depth of any action contained by this action.

        Top-level actions have a depth of zero, and child actions have a depth
        that is one more than their parent action's depth.

        Algorithmically, the notion of max depth is equivalent to the notion of
        height in the context of trees (from graph theory). We decided to use
        this term instead so as not to confuse it with the dimensional height
        of a UI element.

        Returns:
            int: The max depth of any action contained by this action.
        """
        return self._max_depth

    def reset_max_depth(self):
        """Reset the max_depth of this action and all its ancestors to zero."""
        self._max_depth = 0

        if self._parent:
            self._parent.reset_max_depth()

    def render(self, context, action_key='action',
               template_name='reviews/action.html'):
        """Render this action instance and return the content as HTML.

        Args:
            context (django.template.Context):
                The collection of key-value pairs that is passed to the
                template in order to render this action.

            action_key (unicode, optional):
                The key to be used for this action in the context map.

            template_name (unicode, optional):
                The name of the template to be used for rendering this action.

        Returns:
            unicode: The action rendered in HTML.
        """
        content = ''

        if self.should_render(context):
            context.push()

            try:
                context[action_key] = self.copy_to_dict(context)
                content = render_to_string(template_name, context)
            finally:
                context.pop()

        return content

    def register(self, parent=None):
        """Register this review request action instance.

        Note:
           Newly registered top-level actions are appended to the left of the
           other previously registered top-level actions. So if we intend to
           register a collection of top-level actions in a certain order, then
           we likely want to iterate through the actions in reverse.

        Args:
            parent (BaseReviewRequestMenuAction, optional):
                The parent action instance of this action instance.

        Raises:
            KeyError:
                A second registration is attempted (action IDs must be unique
                across all types of actions and menu actions, at any depth).

            DepthLimitExceededError:
                The maximum depth limit is exceeded.
        """
        _populate_defaults()

        if self.action_id in _all_actions:
            raise KeyError('%s already corresponds to a registered review '
                           'request action' % self.action_id)

        if self.max_depth > MAX_DEPTH_LIMIT:
            raise DepthLimitExceededError(self.action_id, MAX_DEPTH_LIMIT)

        if parent:
            parent.child_actions.append(self)
            self._parent = parent
        else:
            _top_level_ids.appendleft(self.action_id)

        _all_actions[self.action_id] = self

    def unregister(self):
        """Unregister this review request action instance.

        Note:
           This method can mutate its parent's child actions. So if we are
           iteratively unregistering a parent's child actions, then we should
           consider first making a clone of the list of children.

        Raises:
            KeyError: An unregistration is attempted before it's registered.
        """
        _populate_defaults()

        try:
            del _all_actions[self.action_id]
        except KeyError:
            raise KeyError('%s does not correspond to a registered review '
                           'request action' % self.action_id)

        if self._parent:
            self._parent.child_actions.remove(self)
        else:
            _top_level_ids.remove(self.action_id)

        self.reset_max_depth()


class BaseReviewRequestMenuAction(BaseReviewRequestAction):
    """A base class for an action with a dropdown menu.

    Note:
        A menu action's child actions must always be pre-registered.
    """

    def __init__(self, child_actions=None):
        """Initialize this menu action.

        Args:
            child_actions (list of BaseReviewRequestAction, optional):
                The list of child actions to be contained by this menu action.

        Raises:
            KeyError:
                A second registration is attempted (action IDs must be unique
                across all types of actions and menu actions, at any depth).

            DepthLimitExceededError:
                The maximum depth limit is exceeded.
        """
        super(BaseReviewRequestMenuAction, self).__init__()

        self.child_actions = []
        child_actions = child_actions or []

        for child_action in child_actions:
            child_action.register(self)

    def copy_to_dict(self, context):
        """Copy this menu action instance to a dictionary.

        Args:
            context (django.template.Context):
                The collection of key-value pairs from the template.

        Returns:
            dict: The corresponding dictionary.
        """
        dict_copy = {
            'child_actions': self.child_actions,
        }
        dict_copy.update(super(BaseReviewRequestMenuAction, self).copy_to_dict(
            context))

        return dict_copy

    @property
    def max_depth(self):
        """Lazily compute the max depth of any action contained by this action.

        Returns:
            int: The max depth of any action contained by this action.
        """
        if self.child_actions and self._max_depth == 0:
            self._max_depth = 1 + max(child_action.max_depth
                                      for child_action in self.child_actions)

        return self._max_depth

    def render(self, context, action_key='menu_action',
               template_name='reviews/menu_action.html'):
        """Render this menu action instance and return the content as HTML.

        Args:
            context (django.template.Context):
                The collection of key-value pairs that is passed to the
                template in order to render this menu action.

            action_key (unicode, optional):
                The key to be used for this menu action in the context map.

            template_name (unicode, optional):
                The name of the template to be used for rendering this menu
                action.

        Returns:
            unicode: The action rendered in HTML.
        """
        return super(BaseReviewRequestMenuAction, self).render(
            context, action_key, template_name)

    def unregister(self):
        """Unregister this review request action instance.

        This menu action recursively unregisters its child action instances.

        Raises:
            KeyError: An unregistration is attempted before it's registered.
        """
        super(BaseReviewRequestMenuAction, self).unregister()

        # Unregistration will mutate self.child_actions, so we make a copy.
        for child_action in list(self.child_actions):
            child_action.unregister()


# TODO: Convert all this to use djblets.registries.
def _populate_defaults():
    """Populate the default action instances."""
    global _populated

    if not _populated:
        _populated = True

        from reviewboard.reviews.default_actions import get_default_actions

        for default_action in reversed(get_default_actions()):
            default_action.register()


def get_top_level_actions():
    """Return a generator of all top-level registered action instances.

    Yields:
        BaseReviewRequestAction:
        All top-level registered review request action instances.
    """
    _populate_defaults()

    return (_all_actions[action_id] for action_id in _top_level_ids)


def register_actions(actions, parent_id=None):
    """Register the given actions as children of the corresponding parent.

    If no parent_id is given, then the actions are assumed to be top-level.

    Args:
        actions (iterable of BaseReviewRequestAction):
            The collection of action instances to be registered.

        parent_id (unicode, optional):
            The action ID of the parent of each action instance to be
            registered.

    Raises:
        KeyError:
            The parent action cannot be found or a second registration is
            attempted (action IDs must be unique across all types of actions
            and menu actions, at any depth).

        DepthLimitExceededError:
            The maximum depth limit is exceeded.
    """
    _populate_defaults()

    if parent_id is None:
        parent = None
    else:
        try:
            parent = _all_actions[parent_id]
        except KeyError:
            raise KeyError('%s does not correspond to a registered review '
                           'request action' % parent_id)

    for action in reversed(actions):
        action.register(parent)

    if parent:
        parent.reset_max_depth()


def unregister_actions(action_ids):
    """Unregister each of the actions corresponding to the given IDs.

    Args:
        action_ids (iterable of unicode):
            The collection of action IDs corresponding to the actions to be
            removed.

    Raises:
        KeyError: An unregistration is attempted before it's registered.
    """
    _populate_defaults()

    for action_id in action_ids:
        try:
            action = _all_actions[action_id]
        except KeyError:
            raise KeyError('%s does not correspond to a registered review '
                           'request action' % action_id)

        action.unregister()


def clear_all_actions():
    """Clear all registered actions.

    This method is really only intended to be used by unit tests. We might be
    able to remove this hack once we convert to djblets.registries.

    Warning:
        This will clear **all** actions, even if they were registered in
        separate extensions.
    """
    global _populated

    _all_actions.clear()
    _top_level_ids.clear()
    _populated = False



from __future__ import unicode_literals

import hashlib
import hmac
import logging
from datetime import datetime

from django.contrib.sites.models import Site
from django.db.models import Model
from django.db.models.query import QuerySet
from django.http.request import HttpRequest
from django.utils import six
from django.utils.safestring import SafeText
from django.utils.six.moves.urllib.parse import (urlencode, urlsplit,
                                                 urlunsplit)
from django.utils.six.moves.urllib.request import (
    HTTPBasicAuthHandler,
    HTTPPasswordMgrWithDefaultRealm,
    Request,
    build_opener)
from django.utils.translation import ugettext as _
from django.template import Context, Template
from django.template.base import Lexer, Parser
from djblets.siteconfig.models import SiteConfiguration
from djblets.webapi.encoders import (BasicAPIEncoder, JSONEncoderAdapter,
                                     ResourceAPIEncoder, XMLEncoderAdapter)

from reviewboard import get_package_version
from reviewboard.notifications.models import WebHookTarget
from reviewboard.reviews.models import Review, ReviewRequest
from reviewboard.reviews.signals import (review_request_closed,
                                         review_request_published,
                                         review_request_reopened,
                                         review_published,
                                         reply_published)


class FakeHTTPRequest(HttpRequest):
    """A fake HttpRequest implementation.

    The WebAPI serialization methods use HttpRequest.build_absolute_uri to
    generate all the links, but none of the various signals that generate
    webhook events have the request plumbed through. Since we don't actually
    need a valid request, this impersonates it enough to get valid results from
    build_absolute_uri.
    """
    _is_secure = None
    _host = None

    def __init__(self, user, local_site_name=None):
        """Initialize a FakeHTTPRequest.

        Args:
            user (django.contrib.auth.models.User):
                The user who initiated the request.

            local_site_name (unicode, optional):
                The local site name (if the request was carried out against a
                local site).
        """
        super(FakeHTTPRequest, self).__init__()

        self.user = user
        self._local_site_name = local_site_name

        if self._is_secure is None:
            siteconfig = SiteConfiguration.objects.get_current()
            self._is_secure = siteconfig.get('site_domain_method') == 'https'
            self._host = Site.objects.get_current().domain

    def is_secure(self):
        return self._is_secure

    def get_host(self):
        return self._host


class CustomPayloadParser(Parser):
    """A custom template parser that blocks certain tags.

    This extends Django's Parser class for template parsing, and removes
    some built-in tags, in order to prevent mailicious use.
    """
    BLACKLISTED_TAGS = ('block', 'debug', 'extends', 'include', 'load', 'ssi')

    def __init__(self, *args, **kwargs):
        super(CustomPayloadParser, self).__init__(*args, **kwargs)

        # Remove some built-in tags that we don't want to expose.
        # There are no built-in filters we have to worry about.
        for tag_name in self.BLACKLISTED_TAGS:
            try:
                del self.tags[tag_name]
            except KeyError:
                pass


def render_custom_content(body, context_data={}):
    """Renders custom content for the payload using Django templating.

    This will take the custom payload content template provided by
    the user and render it using a stripped down version of Django's
    templating system.

    In order to keep the payload safe, we use a limited Context along with a
    custom Parser that blocks certain template tags. This gives us
    tags like {% for %} and {% if %}, but blacklists tags like {% load %}
    and {% include %}.
    """
    lexer = Lexer(body, origin=None)
    parser = CustomPayloadParser(lexer.tokenize())
    template = Template('')
    template.nodelist = parser.parse()

    return template.render(Context(context_data))


def normalize_webhook_payload(payload, request):
    """Normalize a payload for a WebHook, returning a safe, primitive version.

    This will take a payload containing various data types and model references
    and turn it into a payload built out of specific, whitelisted types
    (strings, bools, ints, dicts, lists, and datetimes). This payload is
    safe to include in custom templates without worrying about access to
    dangerous functions, and is easy to serialize.

    Args:
        payload (dict):
            The payload to normalize.

        request (django.http.HttpRequest):
            The HTTP request from the client.

    Returns:
        dict:
        The normalized payload.

    Raises:
        TypeError:
            An unsupported data type was found in the payload. This is an
            issue with the caller.
    """
    def _normalize_key(key):
        if key is None:
            return None

        key_type = type(key)

        if key_type in (bool, int, six.text_type):
            return key

        raise TypeError(
            _('%s is not a valid data type for dictionary keys in '
              'WebHook payloads.')
            % key_type)

    def _normalize_value(value):
        if value is None:
            return None

        value_type = type(value)

        if value_type in (bool, datetime, float, int, long, six.text_type):
            return value
        elif value_type in (SafeText, str):
            return six.text_type(value)
        elif value_type is dict:
            return {
                _normalize_key(dict_key): _normalize_value(dict_value)
                for dict_key, dict_value in six.iteritems(value)
            }
        elif value_type in (list, tuple):
            return [
                _normalize_value(item)
                for item in value
            ]
        elif isinstance(value, (Model, QuerySet)):
            result = resource_encoder.encode(value, request=request)

            if result is not None:
                return _normalize_value(result)

        raise TypeError(
            _('%s is not a valid data type for values in WebHook payloads.')
            % value_type)

    resource_encoder = ResourceAPIEncoder()

    return _normalize_value(payload)


def dispatch_webhook_event(request, webhook_targets, event, payload):
    """Dispatch the given event and payload to the given WebHook targets.

    Args:
        request (django.http.HttpRequest):
            The HTTP request from the client.

        webhook_targets (list of
                         reviewboard.notifications.models.WebHookTarget):
            The list of WebHook targets containing endpoint URLs to dispatch
            to.

        event (unicode):
            The name of the event being dispatched.

        payload (dict):
            The payload data to encode for the WebHook payload.

    Raises:
        ValueError:
            There was an error with the payload format. Details are in the
            log and the exception message.
    """
    try:
        payload = normalize_webhook_payload(payload, request)
    except TypeError as e:
        logging.exception('WebHook payload passed to dispatch_webhook_event '
                          'containing invalid data types: %s',
                          e)

        raise ValueError(six.text_type(e))

    encoder = BasicAPIEncoder()
    bodies = {}

    for webhook_target in webhook_targets:
        if webhook_target.use_custom_content:
            try:
                body = render_custom_content(webhook_target.custom_content,
                                             payload)
                body = body.encode('utf-8')
            except Exception as e:
                logging.exception('Could not render WebHook payload: %s', e)
                continue
        else:
            encoding = webhook_target.encoding

            if encoding not in bodies:
                try:
                    if encoding == webhook_target.ENCODING_JSON:
                        adapter = JSONEncoderAdapter(encoder)
                        body = adapter.encode(payload, request=request)
                        body = body.encode('utf-8')
                    elif encoding == webhook_target.ENCODING_XML:
                        adapter = XMLEncoderAdapter(encoder)
                        body = adapter.encode(payload, request=request)
                    elif encoding == webhook_target.ENCODING_FORM_DATA:
                        adapter = JSONEncoderAdapter(encoder)
                        body = urlencode({
                            'payload': adapter.encode(payload,
                                                      request=request),
                        })
                        body = body.encode('utf-8')
                    else:
                        logging.error('Unexpected WebHookTarget encoding "%s" '
                                      'for ID %s',
                                      encoding, webhook_target.pk)
                        continue
                except Exception as e:
                    logging.exception('Could not encode WebHook payload: %s',
                                      e)
                    continue

                bodies[encoding] = body
            else:
                body = bodies[encoding]

        headers = {
            b'X-ReviewBoard-Event': event.encode('utf-8'),
            b'Content-Type': webhook_target.encoding.encode('utf-8'),
            b'Content-Length': len(body),
            b'User-Agent':
                ('ReviewBoard-WebHook/%s' % get_package_version())
                .encode('utf-8'),
        }

        if webhook_target.secret:
            signer = hmac.new(webhook_target.secret.encode('utf-8'), body,
                              hashlib.sha1)
            headers[b'X-Hub-Signature'] = \
                ('sha1=%s' % signer.hexdigest()).encode('utf-8')

        logging.info('Dispatching webhook for event %s to %s',
                     event, webhook_target.url)
        try:
            url = webhook_target.url
            url_parts = urlsplit(url)

            if url_parts.username or url_parts.password:
                netloc = url_parts.netloc.split('@', 1)[1]
                url = urlunsplit(
                    (url_parts.scheme, netloc, url_parts.path,
                     url_parts.params, url_parts.query))

                password_mgr = HTTPPasswordMgrWithDefaultRealm()
                password_mgr.add_password(
                    None, url, url_parts.username, url_parts.password)
                handler = HTTPBasicAuthHandler(password_mgr)
                opener = build_opener(handler)
            else:
                opener = build_opener()

            opener.open(Request(url.encode('utf-8'), body, headers))
        except Exception as e:
            logging.exception('Could not dispatch WebHook to %s: %s',
                              webhook_target.url, e)


def _serialize_review(review, request):
    return {
        'review_request': review.review_request,
        'review': review,
        'diff_comments': review.comments.all(),
        'file_attachment_comments': review.file_attachment_comments.all(),
        'screenshot_comments': review.screenshot_comments.all(),
        'general_comments': review.general_comments.all(),
    }


def _serialize_reply(reply, request):
    return {
        'review_request': reply.review_request,
        'reply': reply,
        'diff_comments': reply.comments.all(),
        'file_attachment_comments': reply.file_attachment_comments.all(),
        'screenshot_comments': reply.screenshot_comments.all(),
        'general_comments': reply.general_comments.all(),
    }


def review_request_closed_cb(user, review_request, close_type, **kwargs):
    event = 'review_request_closed'
    webhook_targets = WebHookTarget.objects.for_event(
        event, review_request.local_site_id, review_request.repository_id)

    if review_request.local_site_id:
        local_site_name = review_request.local_site.name
    else:
        local_site_name = None

    if webhook_targets:
        if close_type == review_request.SUBMITTED:
            close_type = 'submitted'
        elif close_type == review_request.DISCARDED:
            close_type = 'discarded'
        else:
            logging.error('Unexpected close type %s for review request %s '
                          'when dispatching webhook.',
                          type, review_request.pk)
            return

        if not user:
            user = review_request.submitter

        request = FakeHTTPRequest(user, local_site_name=local_site_name)
        payload = {
            'event': event,
            'closed_by': user,
            'close_type': close_type,
            'review_request': review_request,
        }

        try:
            dispatch_webhook_event(request, webhook_targets, event, payload)
        except ValueError:
            # The error has already been logged. Don't impact the caller.
            pass


def review_request_published_cb(user, review_request, changedesc,
                                **kwargs):
    event = 'review_request_published'
    webhook_targets = WebHookTarget.objects.for_event(
        event, review_request.local_site_id, review_request.repository_id)

    if review_request.local_site_id:
        local_site_name = review_request.local_site.name
    else:
        local_site_name = None

    if webhook_targets:
        request = FakeHTTPRequest(user, local_site_name=local_site_name)
        payload = {
            'event': event,
            'is_new': changedesc is None,
            'review_request': review_request,
        }

        if changedesc:
            payload['change'] = changedesc

        try:
            dispatch_webhook_event(request, webhook_targets, event, payload)
        except ValueError:
            # The error has already been logged. Don't impact the caller.
            pass


def review_request_reopened_cb(user, review_request, **kwargs):
    event = 'review_request_reopened'
    webhook_targets = WebHookTarget.objects.for_event(
        event, review_request.local_site_id, review_request.repository_id)

    if review_request.local_site_id:
        local_site_name = review_request.local_site.name
    else:
        local_site_name = None

    if webhook_targets:
        if not user:
            user = review_request.submitter

        request = FakeHTTPRequest(user, local_site_name=local_site_name)
        payload = {
            'event': event,
            'reopened_by': user,
            'review_request': review_request,
        }

        try:
            dispatch_webhook_event(request, webhook_targets, event, payload)
        except ValueError:
            # The error has already been logged. Don't impact the caller.
            pass


def review_published_cb(user, review, **kwargs):
    event = 'review_published'
    review_request = review.review_request
    webhook_targets = WebHookTarget.objects.for_event(
        event, review_request.local_site_id, review_request.repository_id)

    if review_request.local_site_id:
        local_site_name = review_request.local_site.name
    else:
        local_site_name = None

    if webhook_targets:
        request = FakeHTTPRequest(user, local_site_name=local_site_name)
        payload = _serialize_review(review, request)
        payload['event'] = event

        try:
            dispatch_webhook_event(request, webhook_targets, event, payload)
        except ValueError:
            # The error has already been logged. Don't impact the caller.
            pass


def reply_published_cb(user, reply, **kwargs):
    event = 'reply_published'
    review_request = reply.review_request
    webhook_targets = WebHookTarget.objects.for_event(
        event, review_request.local_site_id, review_request.repository_id)

    if review_request.local_site_id:
        local_site_name = review_request.local_site.name
    else:
        local_site_name = None

    if webhook_targets:
        request = FakeHTTPRequest(user, local_site_name=local_site_name)
        payload = _serialize_reply(reply, request)
        payload['event'] = event

        try:
            dispatch_webhook_event(request, webhook_targets, event, payload)
        except ValueError:
            # The error has already been logged. Don't impact the caller.
            pass


def connect_signals():
    review_request_closed.connect(review_request_closed_cb,
                                  sender=ReviewRequest)
    review_request_published.connect(review_request_published_cb,
                                     sender=ReviewRequest)
    review_request_reopened.connect(review_request_reopened_cb,
                                    sender=ReviewRequest)

    review_published.connect(review_published_cb, sender=Review)
    reply_published.connect(reply_published_cb, sender=Review)



from __future__ import unicode_literals

from django import forms

from reviewboard.hostingsvcs.errors import (AuthorizationError,
                                            TwoFactorAuthCodeRequiredError)
from reviewboard.hostingsvcs.forms import HostingServiceForm
from reviewboard.hostingsvcs.service import HostingService


class TestServiceForm(HostingServiceForm):
    test_repo_name = forms.CharField(
        label='Repository name',
        max_length=64,
        required=True)


class TestService(HostingService):
    name = 'Test Service'
    form = TestServiceForm
    needs_authorization = True
    supports_repositories = True
    supports_bug_trackers = True
    supports_two_factor_auth = True
    has_repository_hook_instructions = True
    supported_scmtools = ['Git']
    bug_tracker_field = ('http://example.com/%(hosting_account_username)s/'
                         '%(test_repo_name)s/issue/%%s')
    repository_fields = {
        'Git': {
            'path': 'http://example.com/%(test_repo_name)s/',
        },
    }

    def authorize(self, username, password, hosting_url, local_site_name=None,
                  two_factor_auth_code=None, *args, **kwargs):
        if username == 'baduser':
            raise AuthorizationError('The username is very very bad.')
        elif username == '2fa-user' and two_factor_auth_code != '123456':
            raise TwoFactorAuthCodeRequiredError('Enter your 2FA code.')

        self.account.data.update({
            'username': username,
            'password': password,
            'hosting_url': hosting_url,
            'local_site_name': local_site_name,
        })

    def is_authorized(self):
        return (self.account.username != 'baduser' and
                'password' in self.account.data)

    def check_repository(self, *args, **kwargs):
        pass


class SelfHostedTestService(TestService):
    name = 'Self-Hosted Test'
    self_hosted = True
    bug_tracker_field = '%(hosting_url)s/%(test_repo_name)s/issue/%%s'
    repository_fields = {
        'Git': {
            'path': '%(hosting_url)s/%(test_repo_name)s/',
            'mirror_path': 'git@%(hosting_domain)s:%(test_repo_name)s/',
        },
    }



"""Unit tests for reviewboard.reviews.views.ReviewFileAttachmentView."""

from __future__ import unicode_literals

from reviewboard.site.urlresolvers import local_site_reverse
from reviewboard.testing import TestCase


class ReviewFileAttachmentViewTests(TestCase):
    """Unit tests for reviewboard.reviews.views.ReviewFileAttachmentView."""

    fixtures = ['test_users']

    def test_access_with_valid_id(self):
        """Testing ReviewFileAttachmentView access with valid attachment for
        review request
        """
        review_request = self.create_review_request(publish=True)
        attachment = self.create_file_attachment(review_request)

        response = self.client.get(
            local_site_reverse(
                'file-attachment',
                kwargs={
                    'review_request_id': review_request.pk,
                    'file_attachment_id': attachment.pk,
                }))
        self.assertEqual(response.status_code, 200)

    def test_access_with_valid_id_and_draft(self):
        """Testing ReviewFileAttachmentView access with valid attachment for
        review request draft
        """
        review_request = self.create_review_request(publish=True)
        attachment = self.create_file_attachment(review_request, draft=True)

        # Log in so that we can check against the draft.
        username = review_request.submitter.username
        self.client.login(username=username, password=username)

        response = self.client.get(
            local_site_reverse(
                'file-attachment',
                kwargs={
                    'review_request_id': review_request.pk,
                    'file_attachment_id': attachment.pk,
                }))
        self.assertEqual(response.status_code, 200)

    def test_access_with_invalid_id(self):
        """Testing ReviewFileAttachmentView access with invalid attachment for
        review request
        """
        review_request = self.create_review_request(publish=True)
        attachment = self.create_file_attachment(review_request)

        review_request2 = self.create_review_request(publish=True)

        response = self.client.get(
            local_site_reverse(
                'file-attachment',
                kwargs={
                    'review_request_id': review_request2.pk,
                    'file_attachment_id': attachment.pk,
                }))
        self.assertEqual(response.status_code, 404)

    def test_access_with_invalid_id_and_draft(self):
        """Testing ReviewFileAttachmentView access with invalid attachment for
        review request draft
        """
        review_request = self.create_review_request(publish=True)
        attachment = self.create_file_attachment(review_request, draft=True)

        review_request2 = self.create_review_request(publish=True)

        # Log in so that we can check against the draft.
        username = review_request.submitter.username
        self.client.login(username=username, password=username)

        response = self.client.get(
            local_site_reverse(
                'file-attachment',
                kwargs={
                    'review_request_id': review_request2.pk,
                    'file_attachment_id': attachment.pk,
                }))
        self.assertEqual(response.status_code, 404)

    def test_access_with_valid_inactive_id(self):
        """Testing ReviewFileAttachmentView access with valid inactive
        attachment for review request
        """
        review_request = self.create_review_request(publish=True)
        attachment = self.create_file_attachment(review_request, active=False)

        response = self.client.get(
            local_site_reverse(
                'file-attachment',
                kwargs={
                    'review_request_id': review_request.pk,
                    'file_attachment_id': attachment.pk,
                }))
        self.assertEqual(response.status_code, 200)

    def test_access_with_valid_inactive_id_draft(self):
        """Testing ReviewFileAttachmentView access with valid inactive
        attachment for review request draft
        """
        review_request = self.create_review_request(publish=True)
        attachment = self.create_file_attachment(review_request, draft=True,
                                                 active=False)

        # Log in so that we can check against the draft.
        username = review_request.submitter.username
        self.client.login(username=username, password=username)

        response = self.client.get(
            local_site_reverse(
                'file-attachment',
                kwargs={
                    'review_request_id': review_request.pk,
                    'file_attachment_id': attachment.pk,
                }))
        self.assertEqual(response.status_code, 200)

    def test_access_with_invalid_inactive_id(self):
        """Testing ReviewFileAttachmentView access with invalid inactive
        attachment for review request
        """
        review_request = self.create_review_request(publish=True)
        attachment = self.create_file_attachment(review_request, active=False)

        review_request2 = self.create_review_request(publish=True)

        response = self.client.get(
            local_site_reverse(
                'file-attachment',
                kwargs={
                    'review_request_id': review_request2.pk,
                    'file_attachment_id': attachment.pk,
                }))
        self.assertEqual(response.status_code, 404)

    def test_access_invalid_inactive_id_draft(self):
        """Testing ReviewFileAttachmentView access with invalid inactive
        attachment for review request draft
        """
        review_request = self.create_review_request(publish=True)
        attachment = self.create_file_attachment(review_request, draft=True,
                                                 active=False)

        review_request2 = self.create_review_request(publish=True)

        # Log in so that we can check against the draft.
        username = review_request.submitter.username
        self.client.login(username=username, password=username)

        response = self.client.get(
            local_site_reverse(
                'file-attachment',
                kwargs={
                    'review_request_id': review_request2.pk,
                    'file_attachment_id': attachment.pk,
                }))
        self.assertEqual(response.status_code, 404)

    def test_access_with_valid_diff_against_id(self):
        """Testing ReviewFileAttachmentView access with valid diff-against
        attachment for review request
        """
        review_request = self.create_review_request(publish=True)
        attachment = self.create_file_attachment(review_request)
        attachment2 = self.create_file_attachment(review_request)

        response = self.client.get(
            local_site_reverse(
                'file-attachment',
                kwargs={
                    'review_request_id': review_request.pk,
                    'file_attachment_id': attachment.pk,
                    'file_attachment_diff_id': attachment2.pk,
                }))
        self.assertEqual(response.status_code, 200)

    def test_access_valid_diff_against_id_draft(self):
        """Testing ReviewFileAttachmentView access with valid diff-against
        attachment for review request draft
        """
        review_request = self.create_review_request(publish=True)
        attachment = self.create_file_attachment(review_request)
        attachment2 = self.create_file_attachment(review_request, draft=True)

        # Log in so that we can check against the draft.
        username = review_request.submitter.username
        self.client.login(username=username, password=username)

        response = self.client.get(
            local_site_reverse(
                'file-attachment',
                kwargs={
                    'review_request_id': review_request.pk,
                    'file_attachment_id': attachment.pk,
                    'file_attachment_diff_id': attachment2.pk,
                }))
        self.assertEqual(response.status_code, 200)

    def test_access_with_invalid_diff_against_id(self):
        """Testing ReviewFileAttachmentView access with invalid diff-against
        attachment for review request
        """
        review_request = self.create_review_request(publish=True)
        attachment = self.create_file_attachment(review_request)

        review_request2 = self.create_review_request(publish=True)
        attachment2 = self.create_file_attachment(review_request2)

        response = self.client.get(
            local_site_reverse(
                'file-attachment',
                kwargs={
                    'review_request_id': review_request.pk,
                    'file_attachment_id': attachment.pk,
                    'file_attachment_diff_id': attachment2.pk,
                }))
        self.assertEqual(response.status_code, 404)

    def test_access_invalid_diff_against_id_draft(self):
        """Testing ReviewFileAttachmentView access with invalid diff-against
        attachment for review request draft
        """
        review_request = self.create_review_request(publish=True)
        attachment = self.create_file_attachment(review_request)

        review_request2 = self.create_review_request(publish=True)
        attachment2 = self.create_file_attachment(review_request2, draft=True)

        # Log in so that we can check against the draft.
        username = review_request.submitter.username
        self.client.login(username=username, password=username)

        response = self.client.get(
            local_site_reverse(
                'file-attachment',
                kwargs={
                    'review_request_id': review_request.pk,
                    'file_attachment_id': attachment.pk,
                    'file_attachment_diff_id': attachment2.pk,
                }))
        self.assertEqual(response.status_code, 404)



"""Gerrit source code hosting support."""

from __future__ import unicode_literals

import base64
import json
import logging

from django import forms
from django.utils.six.moves.urllib.error import HTTPError, URLError
from django.utils.six.moves.urllib.parse import (quote_plus, urlencode,
                                                 urljoin, urlparse)
from django.utils.six.moves.urllib.request import (
    HTTPDigestAuthHandler,
    HTTPPasswordMgrWithDefaultRealm,
    build_opener)
from django.utils.translation import ugettext, ugettext_lazy as _

from reviewboard.hostingsvcs.errors import (AuthorizationError,
                                            HostingServiceError,
                                            RepositoryError,
                                            HostingServiceAPIError)
from reviewboard.hostingsvcs.forms import (HostingServiceAuthForm,
                                           HostingServiceForm)
from reviewboard.hostingsvcs.service import (HostingService,
                                             HostingServiceClient,
                                             URLRequest)
from reviewboard.scmtools.core import Branch, Commit
from reviewboard.scmtools.crypto_utils import (decrypt_password,
                                               encrypt_password)
from reviewboard.scmtools.errors import FileNotFoundError


logger = logging.getLogger(__name__)

_PLUGIN_URL = 'https://github.com/reviewboard/gerrit-reviewboard-plugin/'


class GerritAuthForm(HostingServiceAuthForm):
    """The Gerrit authentication form.

    Gerrit requires an HTTP password to access the web API, so this form saves
    the provided password in the hosting account data.
    """

    def save(self, **kwargs):
        """Save the authentication form.

        Args:
            **kwargs (dict):
                Keyword arguments to pass to
                :py:meth`:HostingServiceAuthForm.save()
                <reviewboard.hostingsvcs.forms.HostingServiceAuthForm.save>`.

        Returns:
            reviewboard.hostingsvcs.models.HostingServiceAccount:
            The hosting service account.

        Raises:
            reviewboard.hostingsvcs.errors.AuthorizationError:
                Information needed to authorize was missing, or authorization
                failed.
        """
        hosting_account = super(GerritAuthForm, self).save(save=False,
                                                           **kwargs)

        if not hosting_account.data:
            hosting_account.data = {}

        hosting_account.data['gerrit_http_password'] = encrypt_password(
            self.get_credentials()['password'])
        hosting_account.save()

        return hosting_account

    class Meta(object):
        help_texts = {
            'hosting_account_username': _(
                'Your HTTP username. You can find this under '
                '<tt>Settings &gt; HTTP Password</tt>.'
            ),
            'hosting_account_password': _(
                'Your HTTP password. You can find this under '
                '<tt>Settings &gt; HTTP Password</tt>.'
            ),
        }


class GerritForm(HostingServiceForm):
    """The Gerrit hosting service form."""

    gerrit_url = forms.URLField(
        label=_('Gerrit URL'),
        required=True,
        help_text=_('The full URL of your Gerrit instance, as in '
                    '<tt>http://gerrit.example.com/</tt>.'),
        widget=forms.TextInput(attrs={'size': 60}))

    gerrit_ssh_port = forms.IntegerField(
        label=_('SSH port'),
        required=True,
        help_text=_('The port configured for SSH access to Gerrit.'),
        initial=29418,
        min_value=1,
        max_value=65535,
        widget=forms.TextInput(attrs={'size': 60}))

    gerrit_project_name = forms.CharField(
        label=_('Project name'),
        required=True,
        help_text=_('The name of the project on Gerrit.'),
        widget=forms.TextInput(attrs={'size': 60}))

    def clean(self):
        """Clean the form.

        This method does no additional validation. It only parses the Gerrit
        URL to determine the domain to use for SSH access (since it will be the
        same domain as for HTTP access).

        Returns:
            dict:
            The cleaned form data.
        """
        self.cleaned_data = super(GerritForm, self).clean()

        gerrit_url = self.cleaned_data.get('gerrit_url')

        if gerrit_url:
            gerrit_domain = urlparse(gerrit_url).netloc

            if ':' in gerrit_domain:
                gerrit_domain = gerrit_domain.split(':', 1)[0]

            self.cleaned_data['gerrit_domain'] = gerrit_domain

        return self.cleaned_data


class GerritClient(HostingServiceClient):
    """The Gerrit hosting service API client."""

    _JSON_PREFIX = ")]}'\n"
    _JSON_PREFIX_LENGTH = len(_JSON_PREFIX)

    def __init__(self, hosting_service):
        """Initialize the client.

        Args:
            hosting_service (reviewboard.hostingsvcs.service.HostingService):
                The hosting service initializing this client.
        """
        super(GerritClient, self).__init__(hosting_service)
        self.account = hosting_service.account

    def get_opener(self, url, username, password):
        """Return an opener with an installed password manager.

        Gerrit's API requires HTTP Digest Auth for its API endpoints. Unlike
        HTTP Basic Auth, this method requires a failed request and response to
        build the digest header, so we cannot pre-compute the header and send
        it along with the initial request.

        Args:
            url (unicode):
                The only part of this required of the URL is the FQDN and the
                URL scheme. The location and fragment portions will be ignored.

            username (unicode):
                The username to use for authentication.

            password (unicode):
                The password to use for authentication.

        Returns:
            urllib2.OpenerDirector:
            The opener with the password manager installed.
        """
        assert username is not None
        assert password is not None

        result = urlparse(url)
        top_level_url = '%s://%s' % (result.scheme, result.netloc)

        password_mgr = HTTPPasswordMgrWithDefaultRealm()
        password_mgr.add_password(None, top_level_url, username, password)
        return build_opener(HTTPDigestAuthHandler(password_mgr))

    def http_request(self, url, body=None, headers=None, method='GET',
                     username=None, password=None):
        """Make an HTTP request to the given URL and return the result.

        This method requires both the username and password arguments to be
        passed since all Gerrit API endpoints require authentication.

        Args:
            url (unicode):
                The URL to make the request against.

            body (unicode, optional):
                The request body.

            headers (dict, optional):
                Additional headers to include in the request.

            method (unicode, optional):
                The HTTP method to use for the request. This defaults to GET.

            username (unicode):
                The username to use for authentication.

            password (unicode):
                The password to use for authentication.

        Returns:
            tuple:
            A 2-tuple of:

            * The response body (:py:class:`bytes`).
            * The response headers (:py:class:`dict`).
        """
        assert username is not None
        assert password is not None

        opener = self.get_opener(url, username, password)
        request = URLRequest(url, body, headers, method=method)

        # Gerrit 2.14+ require Basic Auth, so add that header. Old versions
        # use Digest Auth, which get_opener() already prepared.
        if username is not None and password is not None:
            request.add_basic_auth(username, password)

        try:
            response = opener.open(request)
        except HTTPError as e:
            try:
                raise HostingServiceAPIError(e.reason, e.getcode(), e.read())
            except AttributeError:
                raise HostingServiceError(e.reason, e.getcode())
        except URLError as e:
            raise HostingServiceError(e.reason)

        return response.read(), response.headers

    def api_get(self, url, username=None, password=None, json=True, *args,
                **kwargs):
        """Make a request to the API and return the result.

        Args:
            url (unicode):
                The URL to make the request against.

            username (unicode, optional):
                The username to use when making the request. If not provided,
                the account username will be used.

                This argument must be passed when authorizing.

            password (unicode, optional):
                The password to use when making the request. If not provided,
                the account password will be used.

                This argument must be passed when authorizing.

            json (bool, optional):
                Whether or not to interpret the response as JSON. Defaults to
                ``True``.

            *args (tuple):
                Additional positional arguments to pass to the HTTP request
                method.

            **kwargs (dict):
                Additional keyword arguments to pass to the HTTP request
                method.

        Returns:
            object:
            One of the following:

            * If ``json`` is ``True``, the parsed response (:py:class:`dict`).
            * Otherwise, the raw response (:py:class:`unicode`).
        """
        if json:
            method = self.json_get
        else:
            method = self.http_get

        if username is None or password is None:
            username = self.account.username
            password = decrypt_password(
                self.account.data['gerrit_http_password'])

        return method(
            url,
            username=username,
            password=password,
            *args,
            **kwargs
        )[0]

    def _do_json_method(self, method, url, *args, **kwargs):
        """Make an HTTP request and interpret the result as JSON.

        Gerrit prepends all JSON responses with ``)]}'`` to ensure that they
        cannot be used inline. This method strips that off those characters
        from the response and interprets the rest as JSON.

        Args:
            method (callable):
                The method to use to make the request.

            url (unicode):
                The URL to make the request to

            *args (tuple):
                Positional arguments to pass to ``method``.

            **kwargs (dict):
                Keyword arguments to pass to ``method``.

        Returns:
            tuple:
            A 2-tuple of:

            * The response body parsed as JSON (:py:class:`dict`).
            * The response headers (:py:class:`dict`).
        """
        data, headers = method(url, *args, **kwargs)

        if data:
            if not data.startswith(self._JSON_PREFIX):
                logger.error(
                    'JSON response from Gerrit URL %s does not begin with '
                    '%r: %s',
                    url, self._JSON_PREFIX, data)
            else:
                data = json.loads(data[self._JSON_PREFIX_LENGTH:])

        return data, headers


class Gerrit(HostingService):
    """Source code hosting support for Gerrit.

    Gerrit does not have an API that supports being a hosting service, so the
    gerrit-reviewboard-plugin_ must be installed to use this hosting service.

    If authentication fails to detect the presence of the plugin, the user will
    be informed of such.

    .. _gerrit-reviewboard-plugin:
       https://downloads.reviewboard.org/releases/gerrit-reviewboard-plugin/
    """

    REQUIRED_PLUGIN_VERSION = (1, 0, 0)
    REQUIRED_PLUGIN_VERSION_STR = '%d.%d.%d' % REQUIRED_PLUGIN_VERSION

    name = _('Gerrit')
    client_class = GerritClient
    form = GerritForm
    auth_form = GerritAuthForm
    needs_authorization = True
    supports_repositories = True
    supports_post_commit = True
    supported_scmtools = ['Git']

    repository_fields = {
        'Git': {
            'path': 'ssh://%(hosting_account_username)s@%(gerrit_domain)s:'
                    '%(gerrit_ssh_port)s/%(gerrit_project_name)s',
            'mirror_path':
                'ssh://%(hosting_account_username)s@%(gerrit_domain)s/'
                '%(gerrit_ssh_port)s/%(gerrit_project_name)s',
        },
    }

    def check_repository(self, gerrit_url=None, gerrit_project_name=None,
                         *args, **kwargs):
        """Check that the repository is configured correctly.

        This method ensures that the user has access to an existing repository
        on the Gerrit server and that the ``gerrit-reviewboard`` plugin is
        installed and of a compatible version.

        Args:
            gerrit_url (unicode):
                The URL to the Gerrit server.

            gerrit_project_name (unicode):
                The repository's configured project name on Gerrit.

            *args (tuple):
                Ignored positional arguments.

            **kwargs (dict):
                Ignored keyword arguments.

        Raises:
            reviewboard.hostingsvcs.errors.RepositoryError:
                One of the following:

                * The repository does not exist.
                * The user does not have access to the repository.
                * The ``gerrit-reviewboard`` plugin is not installed.
                * The ``gerrit-reviewboard`` plugin that is installed is an
                  incompatible version.
        """
        url = urljoin(gerrit_url,
                      'a/projects/%s' % quote_plus(gerrit_project_name))

        try:
            self.client.api_get(url, json=False)
        except HostingServiceAPIError as e:
            if e.http_code == 404:
                raise RepositoryError(
                    ugettext('The project "%s" does not exist or you do not '
                             'have access to it.')
                    % gerrit_project_name)

            raise

        url = urljoin(gerrit_url, 'a/plugins/')

        try:
            rsp = self.client.api_get(url)
        except HostingServiceError as e:
            logger.exception(
                'Could not retrieve the list of Gerrit plugins from %s: %s',
                url, e)
            raise RepositoryError(
                ugettext('Could not retrieve the list of Gerrit plugins '
                         'from %(url).')
                % {
                    'url': url,
                })

        if 'gerrit-reviewboard' not in rsp:
            raise RepositoryError(
                ugettext('The "gerrit-reviewboard" plugin is not installed '
                         'on the server. See %(plugin_url)s for installation '
                         'instructions.')
                % {
                    'plugin_url': _PLUGIN_URL,
                })
        else:
            version_str = rsp['gerrit-reviewboard']['version']

            try:
                version = self._parse_plugin_version(version_str)
            except Exception as e:
                logger.exception(
                    'Could not parse gerrit-reviewboard plugin version "%s" '
                    'from %s: %s',
                    version, url, e)
                raise RepositoryError(
                    ugettext('Could not parse gerrit-reviewboard version: '
                             '"%(version)s" from URL %(url)s: %(error)s')
                    % {
                        'version': version,
                        'error': e,
                        'url': url,
                    })

            if version < self.REQUIRED_PLUGIN_VERSION:
                raise RepositoryError(
                    ugettext('The "gerrit-reviewboard" plugin on the server '
                             'is an incompatible version: found %(found)s but '
                             'version %(required)s or higher is required.')
                    % {
                        'found': version_str,
                        'required': self.REQUIRED_PLUGIN_VERSION_STR,
                    }
                )

    def authorize(self, username, password, credentials,
                  local_site_name=None, gerrit_url=None, *args, **kwargs):
        """Authorize against the Gerrit server.

        Args:
            username (unicode):
                The username to use for authentication.

            password  unicode):
                The password to use for authentication.

            credentials (dict):
                The credentials from the authentication form.

            local_site_name (unicode, optional):
                The name of the :py:class:`~reviewboard.site.models.LocalSite`
                the repository is associated with.

            gerrit_url (unicode):
                The URL of the Gerrit server.

            *args (tuple):
                Ignored positional arguments.

            **kwargs (dict):
                Ignored keyword arguments.

        Raises:
            reviewboard.hostingsvcs.errors.AuthorizationError:
                The provided credentials were incorrect.
        """
        if gerrit_url is None:
            raise AuthorizationError('Gerrit URL is required.')

        url = urljoin(gerrit_url, '/a/projects/')

        try:
            self.client.api_get(url, username=username, password=password,
                                json=False)
        except HostingServiceError as e:
            if self.account.pk:
                self.account.data['authorized'] = False

            if e.http_code != 401:
                logger.error(
                    'Unknown HTTP response while authenticating with Gerrit '
                    'at %s: %s',
                    url, e)

                raise AuthorizationError(
                    ugettext('Could not authenticate with Gerrit at %(url): '
                             '%(error)s')
                    % {
                        'url': url,
                        'error': e.message,
                    },
                    http_code=e.http_code)

            raise AuthorizationError(
                ugettext('Unable to authenticate to Gerrit at %(url)s. The '
                         'username or password used may be invalid.')
                % {
                    'url': url,
                },
                http_code=e.http_code)

        self.account.data['authorized'] = True

    def is_authorized(self):
        """Return whether or not the account is authorized with Gerrit.

        Returns:
            bool:
            Whether or not the account has successfully authorized with Gerrit.
        """
        return self.account.data.get('authorized', False)

    def get_branches(self, repository):
        """Return the branches from Gerrit.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository to fetch branches for.

        Returns:
            list of reviewboard.scmtools.core.Branch:
            The list of branches.
        """
        url = self._build_project_api_url(repository, ('branches',))
        rsp = self.client.api_get(url)
        branches = []

        for branch in rsp:
            ref = branch['ref']

            if ref == 'refs/meta/config':
                continue
            elif ref.startswith('refs/heads/'):
                branch_id = ref[len('refs/heads/'):]
                commit = branch['revision']
            else:
                continue

            branches.append(Branch(
                id=branch_id,
                commit=commit,
                default=(branch_id == 'master'),
            ))

        return branches

    def get_commits(self, repository, branch=None, start=None, limit=None):
        """Return a list of commits from the API.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository configured to use Gerrit.

            branch (unicode, optional):
                The branch to retrieve commits for.

            start (unicode, optional):
                The commit SHA1 to start retrieving commits at.

            limit (int, optional):
                The number of commits to retrieve. If unspecified, the default
                is 30 (which is also the maximum).

        Returns:
            list of reviewboard.scmtools.core.Commit:
            The commits returned by the API. These commits will not have diffs
            attached.

        Raises:
            reviewboard.hostingsvcs.errors.HostingServiceAPIError:
                An error occurred communicating with the Gerrit API.
        """
        if not start:
            start = branch

        query = dict(
            (field, bytes(value))
            for field, value in (('start', start), ('limit', limit))
            if value is not None
        )

        url = self._build_project_api_url(repository, ('all-commits',),
                                          query=query)

        try:
            rsp = self.client.api_get(url)
        except HostingServiceAPIError as e:
            # get_change uses this under the hood to retrieve a single commit,
            # so we can specialize the error message to make it more useful in
            # that case.
            if limit == 1:
                raise HostingServiceAPIError(
                    ugettext('Could not retrieve commit "%(rev)s": %(error)s')
                    % {
                        'rev': start,
                        'error': e.message,
                    },
                    http_code=e.http_code,
                    rsp=e.rsp)
            else:
                raise HostingServiceAPIError(
                    ugettext('Could not retrieve commits starting at '
                             '"%(rev)s": %(error)s')
                    % {
                        'rev': start,
                        'error': e.message,
                    },
                    http_code=e.getcode(),
                    rsp=e.rsp)

        return [
            self._parse_commit(meta)
            for meta in rsp
        ]

    def get_file_exists(self, repository, path, revision, *args, **kwargs):
        """Return whether or not the file exists at the specified revision.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository configured to use Gerrit.

            path (unicode):
                The path to the file (ignored).

            revision (unicode):
                The file's Git object ID.

            *args (tuple):
                Ignored positional arguments.

            **kwargs (dict):
                Ignored keyword arguments.

        Returns:
            bool:
            Whether or not the file exists.

        Raises:
            reviewboard.hostingsvcs.errors.HostingServiceAPIError:
                An error occurred communicating with the Gerrit API.
        """
        url = self._build_project_api_url(repository, ('blobs', revision))

        try:
            self.client.api_get(url)
        except HostingServiceAPIError as e:
            if e.http_code == 404:
                return False

            raise

        return True

    def get_file(self, repository, path, revision, *args, **kwargs):
        """Return the contents of a file at the specified revision.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository configured to use Gerrit.

            path (unicode):
                The file path (ignored).

            revision (unicode):
                The file's Git object ID.

            *args (tuple):
                Ignored positional arguments.

            **kwargs (dict):
                Ignored keyword arguments.

        Returns:
            bytes:
            The contents of the file.

        Raises:
            reviewboard.hostingsvcs.errors.FileNotFoundError:
                The file does not exist in the remote repository.

            reviewboard.hostingsvcs.errors.HostingServiceAPIError:
                An error occurred communicating with the Gerrit API.
        """
        url = self._build_project_api_url(repository,
                                          ('blobs', revision, 'content'))

        try:
            rsp = self.client.api_get(url, json=False)
        except HostingServiceAPIError as e:
            if e.http_code == 404:
                raise FileNotFoundError(path, revision=revision)

            raise HostingServiceAPIError(
                ugettext('Could not get file "%(file)s" at revision '
                         '"%(rev)s": %(error)s')
                % {
                    'file': path,
                    'rev': revision,
                    'error': e.read(),
                },
                http_code=e.http_code)

        try:
            return base64.b64decode(rsp)
        except Exception as e:
            raise HostingServiceAPIError(
                ugettext('An error occurred while retrieving "%(file)s" at '
                         'revision "%(rev)s" from Gerrit: the response could '
                         'not be decoded: %(error)s')
                % {
                    'file': path,
                    'rev': revision,
                    'error': e,
                })

    def get_change(self, repository, revision):
        """Return a single change.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository configured to use Gerrit.

            revision (unicode):
                The SHA1 of the commit to fetch.

        Returns:
            reviewboard.scmtools.core.Commit:
            The commit with its diff.
        """
        # The commit endpoint that Gerrit exposes is subtly different from the
        # one our plugin exposes and requires more work on our end to parse, so
        # we just use the all-commits endpoint we expose and limit it to a
        # single commit.
        commit = self.get_commits(repository, start=revision, limit=1)[0]

        url = self._build_project_api_url(repository,
                                          ('commits', revision, 'diff'))
        try:
            diff = self.client.api_get(url, json=False)
        except HostingServiceError as e:
            logger.exception('Could not retrieve change "%s": %s',
                             revision, e)
            raise RepositoryError(
                ugettext('Could not retrieve change "%(rev)s" from repository '
                         '%(repo)d: %(error)s"')
                % {
                    'rev': revision,
                    'repo': repository.id,
                    'error': e,
                })

        commit.diff = diff

        return commit

    def _parse_commit(self, meta, diff=None):
        """Parse and return the meta response and return a Commit.

        Args:
            meta (dict):
                A member of the JSON response from the ``all-commits``
                endpoint corresponding to a single commit.

            diff (bytes, optional):
                The diff corresponding to the commit.

        Returns:
            reviewboard.scmtools.core.Commit:
            The parsed commit.
        """
        commit = Commit(
            author_name=meta['author'],
            id=meta['revision'],
            date=meta['time'],
            message=meta['message'],
            diff=diff
        )

        if meta['parents']:
            commit.parent = meta['parents'][0]

        return commit

    def _parse_plugin_version(self, version_str):
        """Parse a plugin version string into a tuple.

        Args:
            version_str (unicode):
                A version string such as "1.0.0".

        Returns:
            tuple:
            A 3-tuple of the plugin version as ints.
        """
        return tuple(map(int, version_str.split('.')))

    def _build_project_api_url(self, repository, rest_parts, query=None):
        """Return an API URL for the Gerrit projects API.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository configured to use Gerrit.

            rest_parts (iterable):
                The rest of the URL parts.

            **query (dict, optional):
                The query parameters to append to the URL.

        Returns:
            unicode:
            The full URL.
        """
        parts = [
            'a',
            'projects',
            quote_plus(repository.extra_data['gerrit_project_name']),
        ]
        parts.extend(rest_parts)

        url = urljoin(repository.extra_data['gerrit_url'], '/'.join(parts))

        if query:
            url = '%s/?%s' % (url, urlencode(query))
        else:
            url = '%s/' % url

        return url



"""Views for searching."""

from __future__ import unicode_literals

from collections import OrderedDict

from django.contrib.auth.models import User
from django.http import HttpResponseRedirect
from django.shortcuts import render
from django.utils import six
from haystack.generic_views import SearchView

from reviewboard.accounts.mixins import (CheckLoginRequiredViewMixin,
                                         UserProfileRequiredViewMixin)
from reviewboard.avatars import avatar_services
from reviewboard.reviews.models import ReviewRequest
from reviewboard.search import search_backend_registry
from reviewboard.search.forms import RBSearchForm
from reviewboard.site.urlresolvers import local_site_reverse
from reviewboard.site.mixins import CheckLocalSiteAccessViewMixin


class RBSearchView(CheckLoginRequiredViewMixin,
                   CheckLocalSiteAccessViewMixin,
                   UserProfileRequiredViewMixin,
                   SearchView):
    """The Review Board search view."""

    template_name = 'search/results.html'
    disabled_template_name = 'search/search_disabled.html'
    form_class = RBSearchForm

    load_all = False

    ADJACENT_PAGES = 5

    @property
    def paginate_by(self):
        """The number of search results per page."""
        return search_backend_registry.results_per_page

    def dispatch(self, request, local_site=None, *args, **kwargs):
        """Dispatch the view.

        If search is disabled, the search will not be performed and the user
        will be informed.

        Args:
            request (django.http.HttpRequest):
                The current HTTP request.

            local_site (reviewboard.site.models.LocalSite):
                The LocalSite on which the search is being performed.

            *args (tuple, unused):
                Ignored positional arguments.

            **kwargs (dict, unused):
                Ignored keyword arguments.

        Returns:
            django.http.HttpResponse:
            The HTTP response for the search.
        """
        if not search_backend_registry.search_enabled:
            return render(request, self.disabled_template_name)

        form_class = self.get_form_class()
        form = form_class(user=request.user,
                          local_site=local_site,
                          **self.get_form_kwargs())

        if not form.is_valid():
            return self.form_invalid(form)

        query = form.cleaned_data.get(self.search_field, '')

        if not query:
            return HttpResponseRedirect(
                local_site_reverse('all-review-requests',
                                   local_site=local_site),
            )

        if query.isdigit():
            # If the query is an integer, then assume that it's a review
            # request ID that we'll want to redirect to. This mirrors behavior
            # we've had since Review Board 1.7.
            try:
                review_request = ReviewRequest.objects.for_id(query,
                                                              local_site)
            except ReviewRequest.DoesNotExist:
                pass
            else:
                if review_request.is_accessible_by(self.request.user,
                                                   local_site=local_site,
                                                   request=request):
                    return HttpResponseRedirect(
                        review_request.get_absolute_url()
                    )

        return self.form_valid(form)

    def get_context_data(self, form=None, **kwargs):
        """Return context data for rendering the view.

        Args:
            form (reviewboard.search.forms.RBSearchForm):
                The search form instance.

                This will be included in the returned dictionary.

            **kwargs (dict):
                Additional context to be added to the returned dictionary.

        Returns:
            dict:
            The context dictionary.
        """
        context = super(RBSearchView, self).get_context_data(form=form,
                                                             **kwargs)

        paginator = context['paginator']
        page_obj = context['page_obj']
        object_list = context['object_list']

        page_nums = range(max(1, page_obj.number - self.ADJACENT_PAGES),
                          min(paginator.num_pages,
                              page_obj.number + self.ADJACENT_PAGES) + 1)

        active_filters = form.cleaned_data.get('model_filter',
                                               [form.FILTER_ALL])

        context.update({
            'filter_types': OrderedDict(
                (filter_id, dict(active=(filter_id in active_filters),
                                 **filter_type))
                for filter_id, filter_type in six.iteritems(form.FILTER_TYPES)
            ),
            'hits_returned': len(object_list),
            'page_numbers': page_nums,
            'show_first_page': 1 not in page_nums,
            'show_last_page': paginator.num_pages not in page_nums,
        })

        return context

    def render_to_response(self, context, **response_kwargs):
        """Render the search page.

        Args:
            context (dict):
                A dictionary of context from :py:meth:`get_context_data`.

            **response_kwargs (dict);
                Keyword arguments to be passed to the response class.

        Returns:
            django.http.HttpResponse:
            The rendered response.
        """
        show_users = False

        # We only need to fetch users if the search is for just users or
        # both users and review requests (i.e., the '' ID).
        if avatar_services.avatars_enabled:
            show_users = any(
                filter_type['active'] and User in filter_type['models']
                for filter_type in six.itervalues(context['filter_types'])
            )

        if show_users:
            page_obj = context['page_obj']

            user_pks = {
                int(result.pk)
                for result in page_obj
                if result.content_type() == 'auth.user'
            }

            users = {
                user.pk: user
                for user in (
                    User.objects
                    .filter(pk__in=user_pks)
                    .select_related('profile')
                )
            }

            for result in page_obj:
                if result.content_type() == 'auth.user':
                    result.user = users[int(result.pk)]

        return super(RBSearchView, self).render_to_response(context,
                                                            **response_kwargs)



"""Unit tests for reviewboard.accounts.mixins.LoginRequiredViewMixin."""

from __future__ import unicode_literals

from django.contrib.auth.models import AnonymousUser, User
from django.http import HttpResponse, HttpResponseRedirect
from django.test.client import RequestFactory
from django.views.generic.base import View

from reviewboard.accounts.mixins import LoginRequiredViewMixin
from reviewboard.testing import TestCase


class LoginRequiredViewMixinTests(TestCase):
    """Unit tests for reviewboard.accounts.mixins.LoginRequiredViewMixin."""

    def test_dispatch_authenticated_user(self):
        """Testing LoginRequiredViewMixin.dispatch with authenticated user"""
        class MyView(LoginRequiredViewMixin, View):
            def get(view, *args, **kwargs):
                self.assertTrue(view.request.user.is_authenticated())

                return HttpResponse('success')

        request = RequestFactory().request()
        request.user = User.objects.create_user(username='doc',
                                                email='doc@example.com')

        view = MyView.as_view()
        response = view(request)

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, 'success')

    def test_dispatch_anonymous_user(self):
        """Testing LoginRequiredViewMixin.dispatch with anonymous user"""
        class MyView(LoginRequiredViewMixin, View):
            def get(view, *args, **kwargs):
                self.fail('Should not be reached')

        request = RequestFactory().request()
        request.user = AnonymousUser()

        view = MyView.as_view()
        response = view(request)

        self.assertIsInstance(response, HttpResponseRedirect)



"""The search engine backend registry."""

from __future__ import unicode_literals

from djblets.siteconfig.models import SiteConfiguration

from reviewboard.registries.registry import Registry
from reviewboard.search.search_backends.elasticsearch import \
    ElasticsearchBackend
from reviewboard.search.search_backends.whoosh import WhooshBackend


class SearchBackendRegistry(Registry):
    """A registry for search engines backends.

    Extensions can add support for additional search engine backends.

    See :py:ref:`the registry documentation <registry-guides>` for information
    on how registries work.
    """

    lookup_attrs = ['search_backend_id']

    def get_search_backend(self, search_backend_id):
        """Return the search backend with the specified ID.

        Args:
            search_backend_id (unicode):
                The unique identifier of the search engine.

        Returns:
            reviewboard.search.search_engines.base.SearchBackend:
            The search engine class, if it could be found. Otherwise, ``None``.
        """
        return self.get('search_backend_id', search_backend_id)

    def get_defaults(self):
        """Return the default search backends.

        Returns:
            list of reviewboard.search.search_backends.base.SearchBackend:
            The default search backends to use.
        """
        return [
            WhooshBackend(),
            ElasticsearchBackend(),
        ]

    @property
    def on_the_fly_indexing_enabled(self):
        """Whether or not on-the-fly indexing is enabled."""
        siteconfig = SiteConfiguration.objects.get_current()
        return siteconfig.get('search_on_the_fly_indexing')

    @property
    def search_enabled(self):
        """Whether or not search is enabled."""
        siteconfig = SiteConfiguration.objects.get_current()
        return siteconfig.get('search_enable')

    @property
    def results_per_page(self):
        """The number of search results per page."""
        siteconfig = SiteConfiguration.objects.get_current()
        return siteconfig.get('search_results_per_page')

    @property
    def current_backend(self):
        """The current search backend, or ``None`` if search is disabled."""
        if not self.search_enabled:
            return None

        siteconfig = SiteConfiguration.objects.get_current()
        return self.get_search_backend(siteconfig.get('search_backend_id'))



"""Base class for authentication backends."""

from __future__ import unicode_literals

import re
import warnings

from django.contrib.auth.models import User
from djblets.db.query import get_object_or_none

from reviewboard.deprecation import RemovedInReviewBoard40Warning


class BaseAuthBackend(object):
    """Base class for a Review Board authentication backend."""

    #: The unique ID for the authentication backend.
    backend_id = None

    #: The display name for the authentication backend.
    #:
    #: This will be shown in the list of backends in the administration UI.
    name = None

    #: The form class used for authentication settings.
    #:
    #: This must be a subclass of
    #: :py:class:`~djblets.siteconfig.forms.SiteSettingsForm`.
    settings_form = None

    #: Whether this backend supports registering new users.
    supports_registration = False

    #: Whether this backend supports changing the user's full name.
    supports_change_name = False

    #: Whether this backend supports changing the user's e-mail address.
    supports_change_email = False

    #: Whether this backend supports changing the user's password.
    supports_change_password = False

    #: Authentication instructions to display above the Login form.
    login_instructions = None

    #: A regex for matching invalid characters in usernames.
    INVALID_USERNAME_CHAR_REGEX = re.compile(r'[^\w.@+-]')

    def authenticate(self, **credentials):
        """Authenticate a user.

        This will authenticate a user identified by the provided credentials.
        object, or None.

        This must be implemented by subclasses.

        Args:
            **credentials (dict):
                The credentials passed to the backend. This will often
                contain ``username`` and ``password`` keys.

        Returns:
            django.contrib.auth.models.User:
            The authenticated user, or ``None`` if the user could not be
            authenticated for any reason.
        """
        raise NotImplementedError

    def get_or_create_user(self, username, request=None):
        """Return an existing user or create one if it doesn't exist.

        This does not authenticate the user.

        If the user does not exist in the database, but does in the backend,
        its information will be stored in the database for later lookup.
        Subclasses can impose restrictions on this.

        This must be implemented by subclasses.

        Args:
            username (unicode):
                The username to fetch or create.

            request (django.http.HttpRequest, optional):
                The HTTP request from the client.

        Returns:
            django.contrib.auth.models.User:
            The resulting user, or ``None`` if one could not be found.
        """
        raise NotImplementedError

    def get_user(self, user_id):
        """Return an existing user given a numeric user ID.

        Args:
            user_id (int):
                The ID of the user to retrieve.

        Returns:
            django.contrib.auth.models.User:
            The resulting user, or ``None`` if one could not be found.
        """
        return get_object_or_none(User, pk=user_id)

    def update_password(self, user, password):
        """Update a user's password on the backend.

        Authentication backends can override this to update the password
        on the backend. By default, this is not supported.

        Callers should only call this after checking whether
        :py:attr:`supports_change_password` is ``True``.

        Args:
            user (django.contrib.auth.models.User):
                The user whose password will be changed.

            password (unicode):
                The new password.

        Raises:
            NotImplementedError:
                The backend does not support changing passwords.
        """
        raise NotImplementedError

    def update_name(self, user):
        """Update a user's full name on the backend.

        Authentication backends can override this to update the name on the
        backend based on the
        :py:attr:`~django.contrib.auth.models.User.first_name` and
        :py:attr:`~django.contrib.auth.models.User.last_name` values in
        ``user``. By default, this will do nothing.

        Callers should only call this after checking whether
        :py:attr:`supports_change_name` is ``True``.

        Args:
            user (django.contrib.auth.models.User):
                The user whose full name will be changed.
        """
        pass

    def update_email(self, user):
        """Update a user's e-mail address on the backend.

        Authentication backends can override this to update the e-mail
        address on the backend based on the
        :py:attr:`~django.contrib.auth.models.User.email` value in ``user``.
        NBy default, this will do nothing.

        Callers should only call this after checking whether
        :py:attr:`supports_change_email` is ``True``.

        Args:
            user (django.contrib.auth.models.User):
                The user whose full name will be changed.
        """
        pass

    def populate_users(self, query, request, **kwargs):
        """Populate users from the backend into the database based on a query.

        Authentication backends can override this to add users stored on the
        backend into the local database, based on a query string representing
        a full or partial first name, last name, or username. Each result
        that's found based on the query should be stored in the database by
        the backend, generally using the same creation logic as in
        :py:meth:`get_or_create_user`.

        Callers should use this when they need to look up all available users
        from a backend. After calling this, they should look up the results
        from the database.

        If a legacy :py:meth:`query_users` method exists on the class, then
        this will default to calling that with the same parameters (as this was
        the older name for this method). Otherwise, by default, this will do
        nothing.

        Args:
            query (unicode):
                A search query for matching users. This will match the entirety
                or prefix of a username. It's expected that the match will be
                case-insensitive.

            request (django.http.HttpRequest):
                The HTTP request from the client.

            **kwargs (dict):
                Extra positional arguments, for future use.

        Raises:
            reviewboard.accounts.errors.UserQueryError:
                There was an error processing the query or looking up users.
                Details will be in the error message.
        """
        if hasattr(self, 'query_users'):
            warnings.warn('%s.query_users is a deprecated name. Please '
                          'rename it and change the function signature to '
                          'that of query_users().'
                          % self.__class__.__name__,
                          RemovedInReviewBoard40Warning)

            self.query_users(query, request)

    def build_search_users_query(self, query, request, **kwargs):
        """Build a query for searching users in the database.

        This allows backends to construct specialized search queries (
        for use in :py:meth:`QuerySet.filter()
        <django.db.models.query.QuerySet.filter>` when searching for users
        via the :ref:`webapi2.0-user-list-resource`.

        If a legacy :py:meth:`search_users` method exists on the class, then
        this will default to calling that with the same parameters (as this
        was the older name for this method). Otherwise, by default, this will
        return ``None``.

        Args:
            query (unicode):
                A search query for matching users. This will match the entirety
                or prefix of a username. It's expected that the match will be
                case-insensitive.

            request (django.http.HttpRequest):
                The HTTP request from the client.

            **kwargs (dict):
                Extra positional arguments, for future use.

        Returns:
            django.db.models.Q:
            The resulting query for the queryset, or ``None`` to use the
            query for the next available backend (eventually defaulting to
            the standard search query).
        """
        if hasattr(self, 'search_users'):
            warnings.warn('%s.search_users is a deprecated name. Please '
                          'rename it and change the function signature to '
                          'that of build_search_users_query().'
                          % self.__class__.__name__,
                          RemovedInReviewBoard40Warning)

            return self.search_users(query, request)

        return None



from __future__ import unicode_literals

from django_evolution.mutations import AddField, ChangeField
from django.db import models


MUTATIONS = [
    AddField('FileAttachment', 'user', models.ForeignKey, null=True,
             related_model='auth.User'),
    AddField('FileAttachment', 'local_site', models.ForeignKey, null=True,
             related_model='site.LocalSite'),
    ChangeField('FileAttachment', 'file', initial=None, null=True),
]



from __future__ import unicode_literals

from reviewboard.diffviewer.myersdiff import MyersDiffer
from reviewboard.testing import TestCase


class MyersDifferTest(TestCase):
    """Unit tests for MyersDiffer."""

    def test_equals(self):
        """Testing MyersDiffer with equal chunk"""
        self._test_diff(['1', '2', '3'],
                        ['1', '2', '3'],
                        [('equal', 0, 3, 0, 3), ])

    def test_delete(self):
        """Testing MyersDiffer with delete chunk"""
        self._test_diff(['1', '2', '3'],
                        [],
                        [('delete', 0, 3, 0, 0), ])

    def test_insert_before_lines(self):
        """Testing MyersDiffer with insert before existing lines"""
        self._test_diff('1\n2\n3\n',
                        '0\n1\n2\n3\n',
                        [('insert', 0, 0, 0, 2),
                         ('equal', 0, 6, 2, 8)])

    def test_replace_insert_between_lines(self):
        """Testing MyersDiffer with replace and insert between existing lines
        """
        self._test_diff('1\n2\n3\n7\n',
                        '1\n2\n4\n5\n6\n7\n',
                        [('equal', 0, 4, 0, 4),
                         ('replace', 4, 5, 4, 5),
                         ('insert', 5, 5, 5, 9),
                         ('equal', 5, 8, 9, 12)])

    def _test_diff(self, a, b, expected):
        opcodes = list(MyersDiffer(a, b).get_opcodes())
        self.assertEqual(opcodes, expected)



from __future__ import unicode_literals

import json
import logging
import os
from uuid import uuid4

import mimeparse
from django.core.exceptions import ObjectDoesNotExist
from django.http import HttpResponse
from django.template.context import RequestContext
from django.template.loader import render_to_string
from django.utils import six
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext as _

from reviewboard.attachments.mimetypes import MIMETYPE_EXTENSIONS, score_match
from reviewboard.attachments.models import (FileAttachment,
                                            get_latest_file_attachments)
from reviewboard.reviews.context import make_review_request_context
from reviewboard.reviews.markdown_utils import (markdown_render_conditional,
                                                normalize_text_for_edit)
from reviewboard.reviews.models import FileAttachmentComment, Review
from reviewboard.site.urlresolvers import local_site_reverse


_file_attachment_review_uis = []


class ReviewUI(object):
    """Base class for a Review UI.

    Review UIs are interfaces for reviewing content of some type. They take a
    type of object and render a page around it, optionally allowing for the
    display of a diff view for the content. They can render context for
    comments made on the object, provide details for social media sharing (such
    as on a chat or social network).

    A Review UI makes use of a JavaScript side for the interaction, defined
    using :py:attr:`js_model_class` and :py:attr:`js_view_class`. The
    JavaScript side should interface with the API to create/update reviews and
    comments for the object being reviewed.

    Extensions providing Review UIs for file attachments should subclass
    :py:class:`FileAttachmentReviewUI`.

    Attributes:
        diff_against_obj (object):
            The object being diffed against, if any.

        obj (object):
            The object being reviewed.

        request (django.http.HttpRequest):
            The HTTP request from the client. This is only set once
            :py:meth:`render_to_string` is called.

        review_request (reviewboard.reviews.models.review_request.
                        ReviewRequest):
            The review request containing the object being reviewed.
    """

    #: The display name for the Review UI.
    name = None

    #: The template that renders the Review UI.
    #:
    #: Generally, subclasses should use the default template and render the
    #: UI using JavaScript.
    template_name = 'reviews/ui/default.html'

    #: The key passed to the template representing the object.
    object_key = 'obj'

    #: The key passed to the template representing an object to diff against.
    diff_object_key = 'diff_against_obj'

    #: Whether the Review UI can be rendered inline in diffs and other places.
    #:
    #: If set, the Review UI will be able to be displayed within the diff
    #: viewer (and potentially other locations).
    allow_inline = False

    #: Whether this Review UI supports diffing two objects.
    supports_diffing = False

    #: A list of CSS bundle names to include on the Review UI's page.
    css_bundle_names = []

    #: A list of JavaScript bundle names to include on the Review UI's page.
    js_bundle_names = []

    #: A list of specific JavaScript URLs to include on the page.
    #:
    #: It is recommended that :py:attr:`js_bundle_names` be used instead
    #: where possible.
    js_files = []

    #: The name of the JavaScript model class to use for the Review UI.
    js_model_class = None

    #: The name of the JavaScript view class to use for the Review UI.
    js_view_class = None

    def __init__(self, review_request, obj):
        """Initialize the Review UI.

        Args:
            review_request (reviewboard.reviews.models.review_request.
                            ReviewRequest):
                The review request containing the object to review.

            obj (object):
                The object being reviewed.
        """
        self.review_request = review_request
        self.obj = obj
        self.diff_against_obj = None
        self.request = None

    def set_diff_against(self, obj):
        """Set the object to generate a diff against.

        This can only be called on Review UIs that support diffing,
        and must be called before rendering.

        Args:
            obj (object):
                The object being diffed against.
        """
        assert self.supports_diffing

        self.diff_against_obj = obj

    def is_enabled_for(self, user=None, review_request=None, **kwargs):
        """Return whether the Review UI is enabled under the given criteria.

        This can enable or disable a Review UI's functionality depending on
        the user, review request, or some state associated with one or more of
        those.

        When this is called, the arguments are always passed as keyword
        arguments. Subclasses don't need to accept all the arguments, as
        long as they take a ``**kwargs``.

        Args:
            user (django.contrib.auth.models.User, optional):
                The user to check.

            review_request (reviewboard.reviews.models.review_request.
                            ReviewRequest):
                The review request to check.

            **kwargs (dict):
                Additional keyword arguments, for future expansion.

        Returns:
            bool:
            ``True`` if the Review UI is enabled for the given criteria.
            ``False`` otherwise.

            By default, Review UIs are always enabled.
        """
        return True

    def render_to_response(self, request):
        """Render the Review UI to a response.

        This is used to render a page dedicated to the Review UI, complete
        with the standard Review Board chrome.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

        Returns:
            django.http.HttpResponse:
            The HTTP response containing the page for the Review UI.
        """
        return HttpResponse(
            self.render_to_string(request, request.GET.get('inline', False)))

    def render_to_string(self, request, inline=True):
        """Render the Review UI to an HTML string.

        This renders the Review UI to a string for use in embedding into
        either an existing page or a new page.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            inline (bool, optional):
                Whether to render this such that it can be embedded into an
                existing page, instead of as a standalone page.

        Returns:
            django.utils.safestring.SafeText:
            The HTML for the Review UI.
        """
        self.request = request

        try:
            context = self.build_render_context(request, inline=inline)

            return render_to_string(
                self.template_name,
                RequestContext(request, context))
        except Exception as e:
            logging.exception('Error when rendering %r: %s', self, e)

    def build_render_context(self, request, inline, **kwargs):
        """Build context for rendering the page.

        This computes the standard template context to use when rendering the
        page. Generally, subclasses should override
        :py:meth:`get_extra_context`, instead of this.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            inline (bool, optional):
                Whether to render this such that it can be embedded into an
                existing page, instead of as a standalone page.

        Returns:
            dict:
            The context to use in the template.
        """
        last_activity_time, updated_object = \
            self.review_request.get_last_activity()

        draft = self.review_request.get_draft(request.user)
        review_request_details = draft or self.review_request

        close_info = self.review_request.get_close_info()
        caption = self.get_caption(draft)

        context = make_review_request_context(request, self.review_request, {
            'caption': caption,
            'close_description': close_info['close_description'],
            'close_description_rich_text': close_info['is_rich_text'],
            'close_timestamp': close_info['timestamp'],
            'comments': self.get_comments(),
            'draft': draft,
            'last_activity_time': last_activity_time,
            'social_page_image_url': self.get_page_cover_image_url(),
            'social_page_title': (
                'Reviewable for Review Request #%s: %s'
                % (self.review_request.display_id, caption)
            ),
            'review_request_details': review_request_details,
            'review_request': self.review_request,
            'review_ui': self,
            'review_ui_uuid': six.text_type(uuid4()),
            self.object_key: self.obj,
            self.diff_object_key: self.diff_against_obj,
        })

        if inline:
            context.update({
                'base_template': 'reviews/ui/base_inline.html',
                'review_ui_inline': True,
            })
        else:
            context.update({
                'base_template': 'reviews/ui/base.html',
                'review': self.review_request.get_pending_review(request.user),
                'review_ui_inline': False,
            })

        try:
            context.update(self.get_extra_context(request))
        except Exception as e:
            logging.exception('Error when calling get_extra_context for '
                              '%r: %s',
                              self, e)

        return context

    def get_page_cover_image_url(self):
        """Return the URL to an image used to depict this on other sites.

        The returned image URL will be used for services like Facebook, Slack,
        Twitter, etc. when linking to the reviewable object. This may be
        anything from a standard thumbnail to a full-size image.

        By default, no image URL is returned.

        Returns:
            unicode:
            The absolute URL to an image used to depict the reviewable object.
        """
        return None

    def get_comments(self):
        """Return all existing comments on the reviewable object.

        Subclasses must override this.

        Returns:
            list of object:
            The list of comments for the page.
        """
        raise NotImplementedError

    def get_caption(self, draft=None):
        """Return the caption to show for the reviewable object.

        This defaults to requiring ``caption`` and ``draft_caption`` attributes
        on the reviewable object. Subclasses can override this to use something
        else.

        Args:
            draft (reviewboard.reviews.models.review_request_draft.
                   ReviewRequestDraft, optional):
                The active review request draft for the user, if any.

        Returns:
            unicode:
            The caption for the reviewable object.
        """
        raise NotImplementedError

    def get_comment_thumbnail(self, comment):
        """Return an HTML thumbnail for a comment.

        If comment thumbnails are possible for the reviewable object, this
        function should return HTML for the thumbnail.

        Args:
            comment (reviewboard.reviews.models.base_comment.BaseComment):
                The comment to return a thumbnail for.

        Returns:
            django.utils.safestring.SafeText:
            The HTML for a thumbnail for the comment, or ``None`` if one
            can't be generated (using the default thumbnailing for the
            comment type, if one exists).
        """
        return None

    def get_comment_link_url(self, comment):
        """Return a URL for linking to a comment.

        Subclasses must override this.

        Args:
            comment (reviewboard.reviews.models.base_comment.BaseComment):
                The comment to return a link for.

        Returns:
            unicode:
            The URL to link to the comment.
        """
        raise NotImplementedError

    def get_comment_link_text(self, comment):
        """Return the text to link to a comment.

        This must be implemented by subclasses.

        Args:
            comment (reviewboard.reviews.models.base_comment.BaseComment):
                The comment to return text for.

        Returns:
            unicode:
            The text used to link to the comment.
        """
        raise NotImplementedError

    def get_extra_context(self, request):
        """Return extra context to use when rendering the Review UI.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

        Returns:
            dict:
            The context to provide to the template.
        """
        return {}

    def get_js_model_data(self):
        """Return data to pass to the JavaScript Model during instantiation.

        This data will be passed as attributes to the reviewable model
        when constructed.

        Returns:
            dict:
            The attributes to pass to the model.
        """
        return {}

    def get_js_view_data(self):
        """Return data to pass to the JavaScript View during instantiation.

        This data will be passed as options to the reviewable view
        when constructed.

        Returns:
            dict:
            The options to pass to the view.
        """
        return {}

    def get_comments_json(self):
        """Return a JSON-serialized representation of comments for a template.

        The result of this can be used directly in a template to provide
        comments to JavaScript functions.

        Returns:
            unicode:
            Serialized JSON content representing the comments on the reviewable
            object.
        """
        try:
            return mark_safe(json.dumps(
                self.serialize_comments(self.get_comments())))
        except Exception as e:
            logging.exception('Error When calling serialize_comments for '
                              '%r: %s',
                              self, e)

    def serialize_comments(self, comments):
        """Serialize the comments for the Review UI target.

        By default, this will return a list of serialized comments,
        but it can be overridden to return other list or dictionary-based
        representations, such as comments grouped by an identifier or region.
        These representations must be serializable into JSON.

        Args:
            comments (list of reviewboard.reviews.models.base_comment.
                      BaseComment):
                The list of objects to serialize. This will be the result of
                :py:meth:`get_comments`.

        Returns:
            list of dict:
            The list of serialized comment data.
        """
        user = self.request.user

        result = []

        for comment in comments:
            try:
                review = comment.get_review()
            except Review.DoesNotExist:
                logging.error('Missing Review for comment %r' % comment)
                continue

            try:
                if review and (review.public or review.user == user):
                    result.append(self.serialize_comment(comment))
            except Exception as e:
                logging.exception('Error when calling serialize_comment for '
                                  '%r: %s',
                                  self, e)

        return result

    def serialize_comment(self, comment):
        """Serialize a comment.

        This will provide information on the comment that may be useful
        to the JavaScript code.

        Subclasses that want to add additional data should generally
        augment the result of this function and not replace it.

        Args:
            comment (reviewboard.reviews.models.base_comment.BaseComment):
                The comment to serialize.

        Returns:
            dict:
            The serialized comment data.
        """
        review = comment.get_review()
        user = self.request.user

        return {
            'comment_id': comment.pk,
            'text': normalize_text_for_edit(user, comment.text,
                                            comment.rich_text),
            'rich_text': comment.rich_text,
            'html': markdown_render_conditional(comment.text,
                                                comment.rich_text),
            'user': {
                'username': review.user.username,
                'name': review.user.get_profile().get_display_name(user),
            },
            'url': comment.get_review_url(),
            'localdraft': review.user == user and not review.public,
            'review_id': review.pk,
            'review_request_id': review.review_request_id,
            'issue_opened': comment.issue_opened,
            'issue_status': comment.issue_status_to_string(
                comment.issue_status),
        }


class FileAttachmentReviewUI(ReviewUI):
    """Base class for Review UIs for file attachments.

    Review UIs that deal with
    :py:class:`~reviewboard.attachments.models.FileAttachment` objects can
    subclass this to provide the common functionality for their Review UI.

    This class handles fetching and serializing comments, locating a correct
    subclass for a given mimetype, and feeding data to the JavaScript
    :js:class:`RB.AbstractReviewable` model.

    This also handles much of the work for diffing file attachments.
    """

    name = 'Unknown file type'
    object_key = 'file'
    diff_object_key = 'diff_against_file'
    supported_mimetypes = []
    js_model_class = 'RB.DummyReviewable'
    js_view_class = 'RB.DummyReviewableView'

    def is_enabled_for(self, user=None, review_request=None,
                       file_attachment=None, **kwargs):
        """Return whether the Review UI is enabled under the given criteria.

        This can enable or disable a Review UI's functionality, both on the
        file attachment thumbnail and Review UI page, depending on the
        user, review request, file attachment, or some state associated with
        one or more of those.

        When this is called, the arguments are always passed as keyword
        arguments. Subclasses don't need to accept all the arguments, as
        long as they take a ``**kwargs``.

        Args:
            user (django.contrib.auth.models.User, optional):
                The user to check.

            review_request (reviewboard.reviews.models.review_request.
                            ReviewRequest):
                The review request to check.

            file_attachment (reviewboard.attachments.models.FileAttachment):
                The file attachment to check.

            **kwargs (dict):
                Additional keyword arguments, for future expansion.

        Returns:
            bool:
            ``True`` if the Review UI is enabled for the given criteria.
            ``False`` otherwise.

            By default, Review UIs are always enabled.
        """
        return True

    def get_comments(self):
        """Return a list of comments made on the file attachment.

        If this Review UI is showing a diff between two attachments, the
        comments returned will be specific to that diff.

        Returns:
            list of reviewboard.reviews.models.file_attachment_comment.FileAttachmentComment:
            The list of comments on the file attachment or the diff.
        """
        comments = FileAttachmentComment.objects.filter(
            file_attachment_id=self.obj.pk)

        if self.diff_against_obj:
            comments = comments.filter(
                diff_against_file_attachment_id=self.diff_against_obj.pk)
        else:
            comments = comments.filter(
                diff_against_file_attachment_id__isnull=True)

        return comments

    def get_caption(self, draft=None):
        """Return the caption to show for the file attachment.

        Args:
            draft (reviewboard.reviews.models.review_request_draft.
                   ReviewRequestDraft, optional):
                The active review request draft for the user, if any.

        Returns:
            unicode:
            The caption for the file attachment.
        """
        if draft and self.obj.draft_caption:
            return self.obj.draft_caption

        return self.obj.caption

    def get_comment_link_url(self, comment):
        """Return a URL for linking to a comment.

        This will normally just link to the Review UI itself, but subclasses
        may want to override this to provide a more specialized URL.

        Args:
            comment (reviewboard.reviews.models.file_attachment_comment.
                     FileAttachmentComment):
                The comment to return a link for.

        Returns:
            unicode:
            The URL to link to the comment.
        """
        local_site_name = None

        if self.review_request.local_site:
            local_site_name = self.review_request.local_site.name

        return local_site_reverse(
            'file-attachment',
            local_site_name=local_site_name,
            kwargs={
                'review_request_id': self.review_request.display_id,
                'file_attachment_id': self.obj.pk,
            })

    def get_comment_link_text(self, comment):
        """Return the text to link to a comment.

        This will normally just return the file attachment's display name, but
        some may want to specialize to list things like page numbers or
        sections.

        Args:
            comment (reviewboard.reviews.models.file_attachment_comment.
                     FileAttachmentComment):
                The comment to return text for.

        Returns:
            unicode:
            The text used to link to the comment.
        """
        return self.obj.display_name

    def serialize_comment(self, comment):
        """Serialize a comment.

        This will provide information on the comment that may be useful
        to the JavaScript code.

        Args:
            comment (reviewboard.reviews.models.base_comment.BaseComment):
                The comment to serialize.

        Returns:
            dict:
            The serialized comment data.
        """
        data = super(FileAttachmentReviewUI, self).serialize_comment(comment)
        data.update(comment.extra_data)
        return data

    def build_render_context(self, request, inline, **kwargs):
        """Return extra context to use when rendering the Review UI.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            inline (bool, optional):
                Whether to render this such that it can be embedded into an
                existing page, instead of as a standalone page.

        Returns:
            dict:
            The context to use in the template.
        """
        context = super(FileAttachmentReviewUI, self).build_render_context(
            request=request,
            inline=inline)

        context['social_page_title'] = (
            'Attachment for Review Request #%s: %s'
            % (self.review_request.display_id, context['caption'])
        )

        if not inline:
            context['tabs'].append({
                'url': request.path,
                'text': _('File'),
            })

            prev_file_attachment, next_file_attachment = \
                self._get_adjacent_file_attachments(
                    context['review_request_details'])

            context.update({
                'next_file_attachment': next_file_attachment,
                'prev_file_attachment': prev_file_attachment,
            })

        return context

    def get_js_model_data(self):
        """Return model data for the JavaScript AbstractReviewable subclass.

        This will return information on the file attachment, its history,
        and any information on an attachment being diffed against.

        Subclasses can override this to return additional data.

        Returns:
            dict:
            The attributes to pass to the model.
        """
        data = {
            'fileAttachmentID': self.obj.pk,
            'fileRevision': self.obj.attachment_revision,
            'filename': self.obj.orig_filename,
        }

        if self.obj.attachment_history is not None:
            attachments = FileAttachment.objects.filter(
                attachment_history=self.obj.attachment_history)
            data['attachmentRevisionIDs'] = list(
                attachments.order_by('attachment_revision')
                .values_list('pk', flat=True))
            data['numRevisions'] = attachments.count()

        if self.diff_against_obj:
            data['diffCaption'] = self.diff_against_obj.display_name
            data['diffAgainstFileAttachmentID'] = self.diff_against_obj.pk
            data['diffRevision'] = self.diff_against_obj.attachment_revision

            if type(self) != type(self.diff_against_obj.review_ui):
                data['diffTypeMismatch'] = True

        return data

    def _get_adjacent_file_attachments(self, review_request_details):
        """Return the next and previous file attachments.

        The next and previous file attachments are the file attachments that
        occur before and after this one in the review request details view.

        Args:
            review_request_details (reviewboard.reviews.models.
                                    base_review_request_details.
                                    BaseReviewRequestDetails):
                The review request or draft.

        Returns:
            tuple:
            A 2-tuple of the previous and next file attachments, which will
            either be ``None`` (if there isn't a previous or next file
            attachment) or
            :py:class:`~reviewboard.attachments.models.FileAttachment`
            instances.
        """
        file_attachments = iter(get_latest_file_attachments(
            review_request_details.get_file_attachments()))

        prev_obj = None
        next_obj = None

        for obj in file_attachments:
            if obj.pk == self.obj.pk:
                break

            prev_obj = obj

        try:
            next_obj = next(file_attachments)
        except StopIteration:
            pass

        return prev_obj, next_obj

    @classmethod
    def get_best_handler(cls, mimetype):
        """Return the Review UI and score that that best fit the mimetype.

        Args:
            mimetype (unicode):
                The mimetype to find a Review UI for.

        Returns:
            tuple:
            A tuple of ``(best_score, review_ui)``, or ``(0, None)`` if one
            could not be found.
        """
        best_score = 0
        best_fit = None

        for review_ui in _file_attachment_review_uis:
            for mt in review_ui.supported_mimetypes:
                try:
                    score = score_match(mimeparse.parse_mime_type(mt),
                                        mimetype)

                    if score > best_score:
                        best_score = score
                        best_fit = review_ui
                except ValueError:
                    continue

        return best_score, best_fit

    @classmethod
    def for_type(cls, attachment):
        """Return the Review UI that is the best fit for a file attachment.

        Args:
            attachment (reviewboard.attachments.models.FileAttachments):
                The file attachment to locate a Review UI for.

        Returns:
            FileAttachmentReviewUI:
            The Review UI for the attachment, or ``None`` if a suitable one
            could not be found.
        """
        if attachment.mimetype:
            try:
                mimetype = mimeparse.parse_mime_type(attachment.mimetype)
            except:
                logging.error('Unable to parse MIME type "%s" for %s',
                              attachment.mimetype, attachment)
                return None

            # Override the mimetype if mimeparse is known to misinterpret this
            # type of file as 'octet-stream'
            extension = os.path.splitext(attachment.filename)[1]

            if extension in MIMETYPE_EXTENSIONS:
                mimetype = MIMETYPE_EXTENSIONS[extension]

            score, handler = cls.get_best_handler(mimetype)

            if handler:
                try:
                    return handler(attachment.get_review_request(), attachment)
                except ObjectDoesNotExist as e:
                    logging.error('Unable to load review UI for %s: %s',
                                  attachment, e)
                except Exception as e:
                    logging.error('Error instantiating '
                                  'FileAttachmentReviewUI %r: %s',
                                  handler, e)

        return None


def register_ui(review_ui):
    """Register a Review UI class.

    This will register a Review UI. Review Board will use it to display a UI
    when reviewing a supported file attachment.

    Args:
        review_ui (type):
            The Review UI to register. This must be a subclass of
            :py:class:`FileAttachmentReviewUI`.

    Raises:
        TypeError:
            The provided Review UI class is not of a compatible type.
    """
    if not issubclass(review_ui, FileAttachmentReviewUI):
        raise TypeError('Only FileAttachmentReviewUI subclasses can be '
                        'registered')

    _file_attachment_review_uis.append(review_ui)


def unregister_ui(review_ui):
    """Unregister a Review UI class.

    This will unregister a previously registered Review UI.

    Only FileAttachmentReviewUI subclasses are supported. The class must
    have been registered beforehand or a ValueError will be thrown.

    Args:
        review_ui (type):
            The Review UI to unregister. This must be a subclass of
            :py:class:`FileAttachmentReviewUI`, and must have been registered
            before.

    Raises:
        TypeError:
            The provided Review UI class is not of a compatible type.

        ValueError:
            The provided Review UI was not previously registered.
    """
    if not issubclass(review_ui, FileAttachmentReviewUI):
        raise TypeError('Only FileAttachmentReviewUI subclasses can be '
                        'unregistered')

    try:
        _file_attachment_review_uis.remove(review_ui)
    except ValueError:
        logging.error('Failed to unregister missing review UI %r' %
                      review_ui)
        raise ValueError('This review UI was not previously registered')
